
from vsg.rules import split_line_at_token as Rule

from vsg import token

lTokens = []
lTokens.append(token.record_type_definition.end_keyword)


class rule_003(Rule):
    '''
    This rule checks the **end** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic; end record;


    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '003', lTokens)
        self.solution = 'Move *end* keyword and code after it to the next line'




from vsg import violation

from vsg.token import constant_declaration as token
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class rule_007(structure.Rule):
    '''
    This rule checks the **:=** is on the same line at the **constant** keyword.

    **Violation**

    .. code-block:: vhdl

       constant size : integer
          := 1;

    **Fix**

    .. code-block:: vhdl

       constant size : integer := 1;

    **Fix**

    .. code-block:: vhdl

       constant size    : integer := 1;
       constant width   : integer := 32
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'constant', '007')
        self.solution = 'move assignment to same line as constant declaration.'
        self.fixable = False  # Too complicated at the moment to fix.

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by_tokens_if_token_is_between_them(token.constant_keyword, token.semicolon, token.assignment_operator)

    def _analyze(self, lToi):
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
            for oToken in lTokens:
                iLine = utils.increment_line_number(iLine, oToken)

                if isinstance(oToken, token.constant_keyword):
                    iKeywordLine = iLine

                if isinstance(oToken, token.assignment_operator):
                    if iKeywordLine != iLine:
                        oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
                        self.add_violation(oViolation)




from vsg.token import concurrent_conditional_signal_assignment as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import delay_mechanism
from vsg.vhdlFile.classify import conditional_waveforms


def detect(iToken, lObjects):
    '''

    [ label : ] [ postponed ] concurrent_conditional_signal_assignment

    concurrent_conditional_signal_assignment ::=
        target <= [ guarded ] [ delay_mechanism ] conditional_waveforms ;

    conditional_waveforms ::=
        waveform when condition
        { else waveform when condition }
        [ else waveform ]

    The key to detecting this is looking for an assignment <= followed by the keyword **when** before a semicolon.
    '''

    iCurrent = iToken
    bAssignmentFound = False

    while lObjects[iCurrent].get_value() != ';':
        if utils.is_item(lObjects, iCurrent):
            if bAssignmentFound:
                if utils.object_value_is(lObjects, iCurrent, 'when'):
                    return True
            else:
                if utils.object_value_is(lObjects, iCurrent, 'when'):
                    return False
                if utils.object_value_is(lObjects, iCurrent, 'with'):
                    return False

            if utils.object_value_is(lObjects, iCurrent, '<=') and not bAssignmentFound:
                bAssignmentFound = True
        iCurrent += 1

    return False


def classify(iToken, lObjects):
    '''
    concurrent_conditional_signal_assignment ::=
        target <= [ guarded ] [ delay_mechanism ] conditional_waveforms ;
    '''
    iCurrent = utils.assign_tokens_until('<=', token.target, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('<=', token.assignment, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if('guarded', token.guarded_keyword, iCurrent, lObjects)

    iCurrent = delay_mechanism.detect(iCurrent, lObjects)

    iCurrent = conditional_waveforms.classify_until([';'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules.whitespace_before_token import Rule

from vsg.token import conditional_expressions as token


class rule_102(Rule):
    '''
    This rule checks for a single space before the **else** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <= '0' when (rd_en = '0')else '1';

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0' when (rd_en = '0') else '1';
    '''

    def __init__(self):
        Rule.__init__(self, 'conditional_expressions', '102', [token.else_keyword])



import sys
import os
import json
import shutil
import glob
import yaml
import functools
import multiprocessing

from . import junit
from . import rule_list
from . import severity
from . import version
from . import vhdlFile
import argparse
from .CustomArgumentParser import CustomArgumentParser


def __check_strict_positive(value):
    iValue: int = int(value)
    if iValue <= 0:
        raise argparse.ArgumentTypeError(f"{value} is an invalid value for number of jobs")
    return iValue


def __is_valid_file(value: str) -> str:
    """
    Check an argument in argparse to be a path to an existing file
    :param value: String path to analyze.
    :return:
    """
    if not os.path.isfile(value):
        raise argparse.ArgumentTypeError(f"The file {value} does not exist.")
    return value


def parse_command_line_arguments():
    """Parses the command line arguments and returns them."""

    parser = CustomArgumentParser(
        prog='VHDL Style Guide (VSG)',
        description='''Analyzes VHDL files for style guide violations.
                   Reference documentation is located at:
                   http://vhdl-style-guide.readthedocs.io/en/latest/index.html''')

    parser.add_argument('-f', '--filename', type=__is_valid_file, nargs='+', help='File to analyze')
    parser.add_argument('-lr', '--local_rules', help='Path to local rules')
    parser.add_argument('-c', '--configuration', type=__is_valid_file, nargs='+', help='JSON or YAML configuration file(s)')
    parser.add_argument('--fix', default=False, action='store_true', help='Fix issues found')
    parser.add_argument('-fp', '--fix_phase', default=7, action='store',
                        help='Fix issues up to and including this phase')
    parser.add_argument('-j', '--junit', action='store', help='Extract Junit file')
    parser.add_argument('-js', '--json', action='store', help='Extract JSON file')
    parser.add_argument('-of', '--output_format', action='store', default='vsg',
                        choices=['vsg', 'syntastic', 'summary'], help='Sets the output format.')
    parser.add_argument('-b', '--backup', default=False, action='store_true',
                        help='Creates a copy of input file for comparison with fixed version.')
    parser.add_argument('-oc', '--output_configuration', default=None, action='store',
                        help='Write configuration to file name.')
    parser.add_argument('-rc', '--rule_configuration', default=None, action='store',
                        help='Display configuration of a rule')
    parser.add_argument('--style', action='store', default=None, choices=get_predefined_styles(),
                        help='Use predefined style')
    parser.add_argument('-v', '--version', default=False, action='store_true', help='Displays version information')
    parser.add_argument('-ap', '--all_phases', default=False, action='store_true',
                        help='Do not stop when a violation is detected.')
    parser.add_argument('--fix_only', action='store', help='Restrict fixing via JSON file.')
    add_quality_report_argument(parser)
    parser.add_argument(
        "-p",
        "--jobs",
        action="store",
        default=os.cpu_count(),
        type=__check_strict_positive,
        help="number of parallel jobs to use, default is the number of cpu cores",
    )
    parser.add_argument('--debug', default=False, action='store_true', help='Displays verbose debug information')

    args_ = parser.parse_args()

    validate_backup_argument(args_)
    validate_ap_argument(args_)

    if sys.platform == "win32":
        # Work around https://bugs.python.org/issue26903
        args_.jobs = min(args_.jobs, 60)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        return args_


def get_predefined_styles():
    '''
    Reads all predefined styles and returns a list of names.

    Parameters : None

    Returns : (list of strings)
    '''
    lReturn = []
    sStylePath = os.path.join(os.path.dirname(__file__), 'styles')
    lStyles = os.listdir(sStylePath)
    for sStyle in lStyles:
        if sStyle.endswith('.yaml'):
            with open(os.path.join(sStylePath, sStyle)) as yaml_file:
                tempConfiguration = yaml.safe_load(yaml_file)
            lReturn.append(tempConfiguration['name'])
    return lReturn


def validate_backup_argument(args_):
    '''
    The function validates the backup option is only present when the fix option is also present.
    '''
    if args_.backup and not args_.fix:
        print('ERROR:  --backup argument requires --fix argument')
        sys.exit(1)


def validate_ap_argument(args_):
    '''
    The function validates the ap option is not present when the fix option is also present.
    '''
    if args_.all_phases and args_.fix:
        print('ERROR:  -ap argument is invalid with the --fix argument')
        sys.exit(1)


def add_quality_report_argument(parser):
    parser.add_argument(
        '--quality_report',
        action='store',
        help='Create code quality report for GitLab'
    )




from vsg.token import binding_indication as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import entity_aspect
from vsg.vhdlFile.classify import generic_map_aspect
from vsg.vhdlFile.classify import port_map_aspect


def detect(iToken, lObjects):
    if utils.is_next_token('use', iToken, lObjects):
        return classify(iToken, lObjects)
    if utils.is_next_token('generic', iToken, lObjects):
        return classify(iToken, lObjects)
    if utils.is_next_token('port', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):
    '''
    binding_indication ::=
        [ **use** entity_aspect ]
        [ generic_map_aspect ]
        [ port_map_aspect ]
    '''
    iCurrent = iToken

    if utils.is_next_token('use', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('use', token.use_keyword, iCurrent, lObjects)
        iCurrent = entity_aspect.classify(iCurrent, lObjects)

    iCurrent = generic_map_aspect.detect(iCurrent, lObjects)

    iCurrent = port_map_aspect.detect(iCurrent, lObjects)

    return iCurrent




from vsg import parser
from vsg import token

from vsg.rules import consistent_token_case

lTokens = []
lTokens.append(token.variable_declaration.identifier)

lIgnore = []
lIgnore.append(token.interface_signal_declaration.identifier)
lIgnore.append(token.interface_unknown_declaration.identifier)
lIgnore.append(token.interface_constant_declaration.identifier)
lIgnore.append(token.interface_variable_declaration.identifier)
lIgnore.append(token.association_element.formal_part)
lIgnore.append(token.entity_declaration.identifier)
lIgnore.append(token.entity_declaration.entity_simple_name)
lIgnore.append(token.architecture_body.entity_name)
lIgnore.append(parser.whitespace)
lIgnore.append(parser.carriage_return)
lIgnore.append(parser.blank_line)


class rule_011(consistent_token_case):
    '''
    This rule checks for consistent capitalization of variable names.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of entity1 is

         shared variable var1 : std_logic;
         shared variable var2 : std_logic;

       begin

         proc_name : process () is

           variable var3 : std_logic;
           variable var4 : std_logic;

         begin

           Var1 <= '0';

           if (VAR2 = '0') then
             vaR3 <= '1';
           elisif (var2 = '1') then
             VAR4 <= '0';
           end if;

         end process proc_name;

       end architecture rtl;

    **Fix**

    .. code-block:: vhdl

       proc_name : process () is

         variable var1 : std_logic;
         variable var2 : std_logic;
         variable var3 : std_logic;
         variable var4 : std_logic;

       begin

         var1 <= '0';

         if (var2 = '0') then
           var3 <= '1';
         elisif (var2 = '1') then
           var4 <= '0';
         end if;

       end process proc_name;
    '''

    def __init__(self):
        consistent_token_case.__init__(self, 'variable', '011', lTokens, lIgnore)




from .rule_001 import rule_001
from .rule_002 import rule_002
from .rule_003 import rule_003
from .rule_004 import rule_004
from .rule_005 import rule_005
from .rule_006 import rule_006
from .rule_007 import rule_007

from .rule_100 import rule_100
from .rule_101 import rule_101

from .rule_200 import rule_200
from .rule_201 import rule_201

from .rule_300 import rule_300
from .rule_301 import rule_301

from .rule_500 import rule_500
from .rule_501 import rule_501
from .rule_502 import rule_502




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.concurrent_procedure_call_statement.postponed_keyword)


class rule_301(token_indent):
    '''
    This rule checks the indent of the **postponed** keyword if it exists..

    **Violation**

    .. code-block:: vhdl

       a <= b;

         postponed WR_EN(parameter);

    **Fix**

    .. code-block:: vhdl

       a <= b;

       postponed WR_EN(parameter);
    '''

    def __init__(self):
        token_indent.__init__(self, 'procedure_call', '301', lTokens)




from vsg.rules import token_case_with_prefix_suffix

from vsg import token

lTokens = []
lTokens.append(token.process_statement.process_label)


class rule_017(token_case_with_prefix_suffix):
    '''
    This rule checks the process label has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       PROC_A : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'process', '017', lTokens)
        self.groups.append('case::label')




import copy

from vsg import parser
from vsg import token
from vsg import violation

from vsg.rules import alignment_utils
from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class multiline_conditional_alignment(alignment.Rule):
    '''
    This rule checks alignment of multiline concurrent conditional signal statements.

    |configuring_concurrent_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <= '0' when q_wr_en = '1' else
            '1';

       w_foo <= I_FOO when ((I_BAR = '1') and
                (I_CRUFT = '1')) else
                '0';

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0' when q_wr_en = '1' else
                '1';

       w_foo <= I_FOO when ((I_BAR = '1') and
                            (I_CRUFT = '1')) else
                '0';
    '''

    def __init__(self, name, identifier, lTokenPairs):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.subphase = 2
        self.lTokenPairs = lTokenPairs
        self.align_left = 'no'
        self.configuration.append('align_left')
        self.align_paren = 'yes'
        self.configuration.append('align_paren')
        self.wrap_at_when = 'yes'
        self.configuration.append('wrap_at_when')
        self.align_when_keywords = 'no'
        self.configuration.append('align_when_keywords')
        self.align_else_keywords = 'no'
        self.configuration.append('align_else_keywords')

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokenPairs:
            aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            lToi = utils.combine_two_token_class_lists(lToi, aToi)

        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
            iFirstLine, iFirstLineIndent = alignment_utils.get_first_line_info(iLine, oFile)
            oToi.set_meta_data('iFirstLine', iFirstLine)
            oToi.set_meta_data('iFirstLineIndent', iFirstLineIndent)
            oToi.set_meta_data('iAssignColumn', oFile.get_column_of_token_index(oToi.get_start_index()))
            oToi.set_meta_data('bStartsWithParen', _starts_with_paren(lTokens))

        return lToi

    def analyze(self, oFile):
        lToi = self._get_tokens_of_interest(oFile)

        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)

            iFirstLine = oToi.get_meta_data('iFirstLine')
            iFirstLineIndent = oToi.get_meta_data('iFirstLineIndent')
            iAssignColumn = oToi.get_meta_data('iAssignColumn')
            bStartsWithParen = oToi.get_meta_data('bStartsWithParen')

            iColumn = iAssignColumn

            dActualIndent = _build_actual_indent_dict(iLine, lTokens, iFirstLineIndent)

            dIndex = _build_index_dict(iLine, lTokens)

            lActualStructure, iLastLine  = _build_structure_list(iLine, iColumn, lTokens)
            lStructure = copy.deepcopy(lActualStructure)
            if iFirstLine == iLastLine:
                continue

            iFirstIndent = _find_first_indent(self.align_left, dActualIndent, self.indentSize, iAssignColumn)

            dExpectedIndent, lStructure = _apply_align_left_option(self.align_left, lStructure, dActualIndent, bStartsWithParen, self.indentSize, iAssignColumn, iFirstIndent)
            dExpectedIndent, lStructure = _apply_align_paren_option(self.align_paren, lStructure, dExpectedIndent, bStartsWithParen, self.indentSize, iAssignColumn, iFirstIndent)
            dExpectedIndent, lStructure = _apply_align_when_keywords_option(self.align_when_keywords, lStructure, dExpectedIndent, bStartsWithParen, self.indentSize, iAssignColumn, iFirstIndent)
            dExpectedIndent, lStructure = _apply_align_paren_option(self.align_paren, lStructure, dExpectedIndent, bStartsWithParen, self.indentSize, iAssignColumn, iFirstIndent)
            dExpectedIndent, lStructure = _apply_wrap_at_when_option(self.wrap_at_when, lStructure, dExpectedIndent, bStartsWithParen, self.indentSize, iAssignColumn, iFirstIndent)
            if self.wrap_at_when == 'yes' and self.align_paren == 'yes':
                dExpectedIndent, lStructure = _apply_align_paren_after_when(lStructure, dExpectedIndent, bStartsWithParen, self.indentSize, iAssignColumn, iFirstIndent)
            dExpectedIndent, lStructure = _apply_align_else_keywords_option(self.align_else_keywords, lStructure, dExpectedIndent, bStartsWithParen, self.indentSize, iAssignColumn, iFirstIndent)

#            print(lStructure)
#            print(lActualStructure)

#            print(f'Actual = {dActualIndent}')
#            print(f'Expect = {dExpectedIndent}')
#            print(f'Index  = {dIndex}')
#            print(f'dIndex = {dIndex}')

            if self.indentStyle == 'smart_tabs':
                alignment_utils.convert_expected_indent_to_smart_tab(dExpectedIndent, self.indentSize, iFirstLineIndent)

            for iLine in range(iFirstLine + 1, iLastLine + 1):
                if dActualIndent[iLine] == dExpectedIndent[iLine]:
                    continue

                dAction = {}
                dAction['type'] = 'indent'
                dAction['line'] = iLine
                dAction['column'] = dExpectedIndent[iLine]

                if len(dActualIndent[iLine]) > 0:
                    dAction['action'] = 'adjust'
                else:
                    dAction['action'] = 'insert'

                sSolution = alignment_utils.build_solution(dExpectedIndent[iLine])
                iToken = dIndex[iLine]
                oViolation = violation.New(iLine, oToi.extract_tokens(iToken, iToken), sSolution)
                oViolation.set_action(dAction)
                self.add_violation(oViolation)

            if self.align_when_keywords == 'yes':
                for dActual, dExpect in zip(lActualStructure, lStructure):
                    if dActual['type'] == 'when':
                        if dExpect['adjust'] != 0:
                            dAction = {}
                            dAction['type'] = 'when'
                            dAction['line'] = dExpect['line']
                            dAction['column'] = dExpect['column']
                            dAction['adjust'] = dExpect['adjust']
                            sSolution = 'Align when with other whens at column ' + str(dExpect['column'])
                            oViolation = violation.New(dAction['line'], oToi.extract_tokens(dActual['iToken'] - 1, dActual['iToken'] - 1), sSolution)
                            oViolation.set_action(dAction)
                            self.add_violation(oViolation)

            if self.align_else_keywords == 'yes':
                for dActual, dExpect in zip(lActualStructure, lStructure):
                    if dActual['type'] == 'else':
                        if dExpect['adjust'] != 0:
                            dAction = {}
                            dAction['type'] = 'else'
                            dAction['line'] = dExpect['line']
                            dAction['column'] = dExpect['column']
                            dAction['adjust'] = dExpect['adjust']
                            sSolution = 'Align else with other elses at column ' + str(dExpect['column'])
                            oViolation = violation.New(dAction['line'], oToi.extract_tokens(dActual['iToken'] - 1, dActual['iToken'] - 1), sSolution)
                            oViolation.set_action(dAction)
                            self.add_violation(oViolation)

        self._sort_violations()

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
#        print(dAction)
        if dAction['type'] == 'when':
            iSpace = len(lTokens[0].get_value())
            iNewSpace = iSpace + dAction['adjust']
            lTokens[0].set_value(' '*iNewSpace)
        elif dAction['type'] == 'else':
            iSpace = len(lTokens[0].get_value())
            iNewSpace = iSpace + dAction['adjust']
            lTokens[0].set_value(' '*iNewSpace)
        elif dAction['type'] == 'indent':
            if dAction['action'] == 'adjust':
                lTokens[0].set_value(dAction['column'])
            else:
                rules_utils.insert_new_whitespace(lTokens, 0, dAction['column'])

        oViolation.set_tokens(lTokens)


def _find_first_column(oFile, oToi, sAlignLeft, iIndentSize, iIndentStep):
    iStartIndex = oToi.get_start_index()

    if sAlignLeft == 'yes':
        iIndentLevel = oFile.get_indent_of_line_at_index(iStartIndex)
        iFirstColumn = iIndentLevel * iIndentSize + iIndentStep
    else:
        iFirstColumn = oFile.get_column_of_token_index(iStartIndex)

    return iFirstColumn


def is_token_before_carriage_return(tToken, lTokens):
    for oToken in lTokens:
        if isinstance(oToken, tToken):
            return True
        if isinstance(oToken, parser.carriage_return):
            return False
    return False


def _apply_align_left_option(sConfig, lStructure, dActualIndent, bStartsWithParen, iIndentStep, iAssignColumn, iFirstIndent):
#    print('--> _apply_align_left_option  <-' + '-'*70)
    iFirstLine = alignment_utils.get_first_line(dActualIndent)

    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

    bWhenFound = False
    iParens = 0
    iLine = iFirstLine
    for dStruct in lStructure:
        if dStruct['type'] == 'when':
            bWhenFound = True
        elif dStruct['type'] == 'else':
            bWhenFound = False
        elif dStruct['type'] == 'return':
#            print(f'iLine = {iLine} | bWhenFound = {bWhenFound} | iParens = {iParens}')
            iLine += 1
            if bWhenFound:
                if iLine == iFirstLine:
                    iIndent = iFirstIndent
                    iIndent += iIndentStep
                elif iLine == iFirstLine + 1:
                    iIndent = iFirstIndent
                    iIndent += iIndentStep
                elif len(dExpectedIndent[iLine - 1]) == iFirstIndent:
                    iIndent = iFirstIndent
                    iIndent += iIndentStep
                else:
                    iIndent = len(dExpectedIndent[iLine - 1])
#                print(f'[iIndent]{iIndent}[iParens]{iParens}[iIndentStep]{iIndentStep}')
                iIndent += iParens * iIndentStep
            else:
                iIndent = iFirstIndent
            dExpectedIndent[iLine] = iIndent * ' '
        elif dStruct['type'] == 'open':
            iParens +=1
        elif dStruct['type'] == 'close':
            iParens -= 1

    lReturnStructure = _update_structure(dExpectedIndent, dActualIndent, lStructure)
    return dExpectedIndent, lReturnStructure


def _apply_align_paren_option(sConfig, lStructure, dActualIndent, bStartsWithParen, iIndentStep, iAssignColumn, iFirstIndent):
#    print('--> _apply_align_paren_option <-' + '-'*70)
    if sConfig == 'no':
        return dActualIndent, lStructure
    iFirstLine = alignment_utils.get_first_line(dActualIndent)
    iLastLine = alignment_utils.get_last_line(dActualIndent)

    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

#    if bStartsWithParen:
#       iIndent = dActualIndent[iFirstLine]
#       iColumn = iIndent
#       lColumn = [dActualIndent[iFirstLine]]
#    else:
    iIndent = iFirstIndent
    iColumn = iIndent
    lColumn = [iIndent]

#    print('*'*80)
#    print(lParens)

    bWhenFound = False
    iParens = 0
    lWhenIndent = []
    lParens = []
    for dStruct in lStructure:
        if dStruct['type'] == 'when':
            bWhenFound = True
        elif dStruct['type'] == 'else':
            bWhenFound = False
        elif dStruct['type'] == 'return' and bWhenFound:
            if iParens == 0:
                lWhenIndent.append(dStruct['line'])
        elif dStruct['type'] == 'open':
            iParens +=1
            lParens.append(dStruct)
        elif dStruct['type'] == 'close':
            iParens -= 1
            lParens.append(dStruct)

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')
        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

#        print(f'lTemp = {lTemp}')
        iTemp = lColumn[-1]
        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
                if iLine == iFirstLine:
                    iColumn = dTemp['column'] + iIndentStep - 1
                else:
                    iColumn = dTemp['column'] + (iTemp - len(dActualIndent[iLine])) + iIndentStep - 1
#                print(f"iColumn = {dTemp['column']} + ({iTemp} - {dActualIndent[iLine]}) + {iIndentStep} - 1 = {iColumn}")
                lColumn.append(iColumn)
                dExpectedIndent[iLine + 1] = iColumn * ' '
            else:
                iParens -= 1
                lColumn.pop()
                dExpectedIndent[iLine + 1] = lColumn[-1] * ' '
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] = (len(dExpectedIndent[iLine]) - iIndentStep) * ' '

#        print(f'iParens = {iParens}')


        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = lColumn[-1] * ' '

#        print(f'{iLine} | {lColumn} | {dExpectedIndent}')

        if (iLine + 1) in lWhenIndent:
            dExpectedIndent[iLine + 1] = (iFirstIndent + iIndentStep) * ' '

    lReturnStructure = _update_structure(dExpectedIndent, dActualIndent, lStructure)
    return dExpectedIndent, lReturnStructure


def _apply_align_when_keywords_option(sConfig, lStructure, dActualIndent, bStartsWithParen, iIndentStep, iAssignColumn, iFirstIndent):
#    print('--> _apply_align_when_keywords_option <-' + '-'*70)
    if sConfig == 'no':
        return dActualIndent, lStructure

    iWhenMax = -1
    for dStruct in lStructure:
        if dStruct['type'] == 'when':
            iWhenMax = max(iWhenMax, dStruct['column'])

#    print(f'iWhenMax = {iWhenMax}')

    lNewStruct = []
    bAdjust = False
    for dStruct in lStructure:
        if dStruct['type'] == 'when':
            iAdjust = iWhenMax - dStruct['column']
            if iAdjust != 0:
                bAdjust = True
            dStruct['adjust'] = iAdjust
        if dStruct['type'] == 'return':
            bAdjust = False
        if bAdjust:
            dStruct['column'] += iAdjust
        lNewStruct.append(dStruct)

    return dActualIndent, lNewStruct


def _apply_wrap_at_when_option(sConfig, lStructure, dActualIndent, bStartsWithParen, iIndentStep, iAssignColumn, iFirstIndent):
#    print('--> _apply_wrap_at_when_option <-' + '-'*70)
    if sConfig == 'no':
        return dActualIndent, lStructure
    iFirstLine = alignment_utils.get_first_line(dActualIndent)
    iLastLine = alignment_utils.get_last_line(dActualIndent)

    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

    bWhenFound = False
    dWhenIndent = {}
    lParens = []
    for dStruct in lStructure:
        if dStruct['type'] == 'when':
            bWhenFound = True
            iColumn = dStruct['column']
        elif dStruct['type'] == 'else':
            bWhenFound = False
        elif dStruct['type'] == 'return' and bWhenFound:
            dTemp = {}
            dTemp['column'] = iColumn
            dWhenIndent[dStruct['line']] = dTemp
        elif dStruct['type'] == 'open':
            lParens.append(dStruct)
        elif dStruct['type'] == 'close':
            lParens.append(dStruct)

    iWhenIndent = -1
    iParens = 0

    for iLine in range(iFirstLine, iLastLine):

        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
            else:
                iParens -= 1

        if iLine + 1 in dWhenIndent:
            if iLine in dWhenIndent:
                iWhenIndent = len(dExpectedIndent[iLine]) + (iParens * iIndentStep)
            else:
                iWhenIndent = dWhenIndent[iLine + 1]['column'] + 4 + 1 + (iParens * iIndentStep)
            dExpectedIndent[iLine + 1] = iWhenIndent * ' '
        else:
            dExpectedIndent[iLine + 1] = dActualIndent[iLine + 1]

    lReturnStructure = _update_structure(dExpectedIndent, dActualIndent, lStructure)
    return dExpectedIndent, lReturnStructure


def _apply_align_paren_after_when(lStructure, dActualIndent, bStartsWithParen, iIndentStep, iAssignColumn, iFirstIndent):
#    print('--> _apply_align_paren_option <-' + '-'*70)
    iFirstLine = alignment_utils.get_first_line(dActualIndent)
    iLastLine = alignment_utils.get_last_line(dActualIndent)

    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

#    if bStartsWithParen:
#       iIndent = dActualIndent[iFirstLine]
#       iColumn = iIndent
#       lColumn = [dActualIndent[iFirstLine]]
#    else:
    iIndent = iFirstIndent
    iColumn = iIndent
    lColumn = [iIndent]

#    print('*'*80)
#    print(lParens)

    bWhenFound = False
    iParens = 0
    lWhenIndent = []
    lParens = []
    for dStruct in lStructure:
        if dStruct['type'] == 'when':
            bWhenFound = True
        elif dStruct['type'] == 'else':
            bWhenFound = False
        elif dStruct['type'] == 'return' and bWhenFound:
            if iParens == 0:
                lWhenIndent.append(dStruct['line'])
        elif dStruct['type'] == 'open':
            iParens +=1
            lParens.append(dStruct)
        elif dStruct['type'] == 'close':
            iParens -= 1
            lParens.append(dStruct)

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')
        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

#        print(f'lTemp = {lTemp}')
        iTemp = lColumn[-1]
        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
                if iLine == iFirstLine:
                    iColumn = dTemp['column'] + iIndentStep - 1
                else:
                    iColumn = dTemp['column'] + (iTemp - len(dActualIndent[iLine])) + iIndentStep - 1
#                print(f"iColumn = {dTemp['column']} + ({iTemp} - {dActualIndent[iLine]}) + {iIndentStep} - 1 = {iColumn}")
                lColumn.append(iColumn)
                dExpectedIndent[iLine + 1] = iColumn * ' '
            else:
                iParens -= 1
                lColumn.pop()
                dExpectedIndent[iLine + 1] = lColumn[-1] * ' '
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] = (len(dExpectedIndent[iLine]) - iIndentStep) * ' '

#        print(f'iParens = {iParens}')


        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = lColumn[-1] * ' '

#        print(f'{iLine} | {lColumn} | {dExpectedIndent}')

        if (iLine + 1) in lWhenIndent and iParens == 0:
            dExpectedIndent[iLine + 1] = dActualIndent[iLine + 1]

    lReturnStructure = _update_structure(dExpectedIndent, dActualIndent, lStructure)
    return dExpectedIndent, lReturnStructure


def _apply_align_else_keywords_option(sConfig, lStructure, dActualIndent, bStartsWithParen, iIndentStep, iAssignColumn, iFirstIndent):
#    print('--> _apply_align_when_keywords_option <-' + '-'*70)

    if sConfig == 'no':
        return dActualIndent, lStructure

    iElseMax = -1
    for dStruct in lStructure:
        if dStruct['type'] == 'else':
            iElseMax = max(iElseMax, dStruct['column'])

#    print(f'iElseMax = {iElseMax}')

    lNewStruct = []
    bAdjust = False
    for dStruct in lStructure:
        if dStruct['type'] == 'else':
            iAdjust = iElseMax - dStruct['column']
#            print(f'{iAdjust} = {iElseMax} - {dStruct["column"]}')
            if iAdjust != 0:
                bAdjust = True
            dStruct['adjust'] = iAdjust
        if dStruct['type'] == 'return':
            bAdjust = False
        if bAdjust:
            dStruct['column'] += iAdjust
        lNewStruct.append(dStruct)

    return dActualIndent, lNewStruct


def _starts_with_paren(lTokens):

    iToken = utils.find_next_non_whitespace_token(1, lTokens)
    if isinstance(lTokens[iToken], parser.open_parenthesis):
        return True
    return False


def _build_actual_indent_dict(iLine, lTokens, iFirstLineIndent):
    dReturn = {}
    dReturn[iLine] = iFirstLineIndent * ' '

    for iToken, oToken in enumerate(lTokens):

        iLine = utils.increment_line_number(iLine, oToken)

        if isinstance(oToken, parser.blank_line):
            continue

        if isinstance(oToken, parser.carriage_return):
            dReturn[iLine] = alignment_utils.set_indent(iToken, lTokens)
            continue

    return dReturn


def _build_index_dict(iLine, lTokens):
    dReturn = {}

    for iToken, oToken in enumerate(lTokens):

        iLine = utils.increment_line_number(iLine, oToken)

        if isinstance(oToken, parser.blank_line):
            continue

        if isinstance(oToken, parser.carriage_return):
            dReturn[iLine] = iToken + 1
            continue

    return dReturn


def _build_structure_list(iLine, iColumn, lTokens):

    lStructure = []

    for iToken, oToken in enumerate(lTokens):

        iLine = utils.increment_line_number(iLine, oToken)

        if isinstance(oToken, parser.blank_line):
            continue

        if isinstance(oToken, parser.carriage_return):
            iColumn = 0
            dReturn = {}
            dReturn['type'] = 'return'
            dReturn['line'] = iLine
            lStructure.append(dReturn)
            continue

        iColumn += len(oToken.get_value())

        if isinstance(oToken, parser.close_parenthesis):
            dParen = {}
            dParen['type'] = 'close'
            dParen['line'] = iLine
            dParen['column'] = iColumn
            dParen['begin_line'] = utils.does_token_start_line(iToken, lTokens)
            lStructure.append(dParen)

        if isinstance(oToken, parser.open_parenthesis):
            dParen = {}
            dParen['type'] = 'open'
            dParen['line'] = iLine
            dParen['column'] = iColumn
            lStructure.append(dParen)

        if oToken.get_value().lower() == 'when':
            dWhen = {}
            dWhen['type'] = 'when'
            dWhen['line'] = iLine
            dWhen['column'] = iColumn - 4
            dWhen['iToken'] = iToken
            lStructure.append(dWhen)

        if oToken.get_value().lower() == 'else':
            dElse = {}
            dElse['type'] = 'else'
            dElse['line'] = iLine
            dElse['column'] = iColumn - 4
            dElse['iToken'] = iToken
            lStructure.append(dElse)

    return lStructure, iLine


def _update_structure(dExpectedIndent, dActualIndent, lStructure):
#    print('--> _update_structure <' + '-'*80)
    iFirstLine = alignment_utils.get_first_line(dActualIndent)
    iLastLine = alignment_utils.get_last_line(dActualIndent)
    lReturn = []
#    print(f'dActaulIndent = {dActualIndent}')
#    print(f'dExpectedIndent = {dExpectedIndent}')
    for iLine in range(iFirstLine, iLastLine + 1):
#        print(f'dActualIndent = |{dActualIndent[iLine]}|')
#        print(f'dExpectedIndent = |{dExpectedIndent[iLine]}|')
        iDeltaIndent = len(dExpectedIndent[iLine]) - len(dActualIndent[iLine])
        for dStruct in lStructure:
            if dStruct['line'] == iLine:
                if dStruct['type'] != 'return':
                    dStruct['column'] += iDeltaIndent
                lReturn.append(dStruct)
    return lReturn


def _find_first_indent(sConfig, dActualIndent, iIndentStep, iAssignColumn):
    iFirstLine = alignment_utils.get_first_line(dActualIndent)
    if sConfig == 'yes':
        iFirstIndent = len(dActualIndent[iFirstLine]) + iIndentStep
    else:
        iFirstIndent = iAssignColumn + 2 + 1
    return iFirstIndent




from vsg import parser

def get_column_of_token_index(iToken, lAllTokens, oTokenMap):

    oToken = lAllTokens[iToken]
    iLine = oTokenMap.get_line_number_of_index(iToken) - 1
    lCarriageReturns = oTokenMap.get_token_indexes(parser.carriage_return)
    iStart = lCarriageReturns[iLine - 1] + 1

    iReturn = 0
    for oToken in lAllTokens[iStart:iToken]:
        iReturn += len(oToken.get_value())
    return iReturn




from vsg import parser

from vsg.vhdlFile.extract import tokens


def get_tokens_matching_not_at_beginning_or_ending_of_line(lTokens, lAllTokens, oTokenMap):

    lReturn = []
    lIndexes = []
    lCarriageReturns = oTokenMap.get_token_indexes(parser.carriage_return)

    for oToken in lTokens:
        lIndexes.extend(oTokenMap.get_token_indexes(oToken))

    lIndexes.sort()

    for iIndex in lIndexes:
        if iIndex - 1 not in lCarriageReturns and iIndex + 1 not in lCarriageReturns:
            iLine = oTokenMap.get_line_number_of_index(iIndex)
            lReturn.append(tokens.New(iIndex, iLine, [lAllTokens[iIndex]]))

    return lReturn





from vsg import parser
from vsg import violation

from vsg.rule_group import blank_line
from vsg.vhdlFile import utils


class blank_lines_between_token_pairs(blank_line.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token type pairs
       The tokens to check for a single space between
    '''

    def __init__(self, name, identifier, lTokenPairs):
        blank_line.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = 'Remove blank line'
        self.lTokenPairs = lTokenPairs

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokenPairs:
            lToi_a = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            lToi = utils.combine_two_token_class_lists(lToi, lToi_a)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)

            for iToken, oToken in enumerate(lTokens):
                iLine = utils.increment_line_number(iLine, oToken)

                if isinstance(oToken, parser.blank_line):
                    oNewToi = oToi.extract_tokens(iToken, iToken + 1)
                    oViolation = violation.New(iLine, oNewToi, self.solution)
                    self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        oViolation.set_tokens([])




from vsg import deprecated_rule


class rule_007(deprecated_rule.Rule):
    '''
    This rule has been renamed to `port_map_004 <port_map_rules.html#port-map-004>`_.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'instantiation', '007')
        self.message.append('Rule ' + self.unique_id + ' has been renamed to port_map_004.')




from vsg import token

from vsg.rules import move_token_left_to_next_non_whitespace_token as Rule

oToken = token.entity_declaration.semicolon


class rule_025(Rule):
    '''
    This rule checks the semicolon is not on its own line.

    **Violation**

    .. code-block:: vhdl

       end entity;

       end entity
       ;

    **Fix**

    .. code-block:: vhdl

       end entity;

       end entity;
    '''

    def __init__(self):
        Rule.__init__(self, 'entity', '025', oToken)
        self.bInsertWhitespace = False
        self.subphase = 3




from vsg.token import generic_clause as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import generic_list


def detect(iToken, lObjects):
    '''
    generic_clause ::=
        generic ( generic_list ) ;
    '''
    if utils.are_next_consecutive_tokens(['generic', '('], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('generic', token.generic_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iCurrent, lObjects)

    iCurrent = generic_list.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import parser


class postponed_keyword(parser.keyword):
    '''
    unique_id = concurrent_signal_assignment_statement : postponed_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)

class label_name(parser.label):
    '''
    unique_id = concurrent_signal_assignment_statement : label_name
    '''

    def __init__(self, sString):
        parser.label.__init__(self, sString)

class label_colon(parser.colon):
    '''
    unique_id = concurrent_signal_assignment_statement : label_colon
    '''

    def __init__(self, sString=';'):
        parser.colon.__init__(self, sString)




from vsg import parser

from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_tokens_at_beginning_of_line_matching(lTokens, lAllTokens, oTokenMap):

    lIndexes = utils.get_indexes_of_token_list(lTokens, oTokenMap)
    lReturn = []
    for iIndex in lIndexes:
        if oTokenMap.is_token_at_index(parser.carriage_return, iIndex - 1):
            iLine = oTokenMap.get_line_number_of_index(iIndex)
            lReturn.append(tokens.New(iIndex, iLine, [lAllTokens[iIndex]]))
        elif oTokenMap.is_token_at_index(parser.carriage_return, iIndex - 2) and utils.is_token_at_index_whitespace(oTokenMap, iIndex - 1):
            iLine = oTokenMap.get_line_number_of_index(iIndex)
            lReturn.append(tokens.New(iIndex - 1, iLine, lAllTokens[iIndex - 1: iIndex + 1]))
    return lReturn




from vsg import rule


class Rule(rule.Rule):
    '''
    Class for assigning rules to the naming group.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.
    '''

    def __init__(self, name, identifier):
        rule.Rule.__init__(self, name=name, identifier=identifier)
        self.phase = 7
        self.groups.append('naming')




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.if_statement.elsif_keyword)


class rule_026(token_case):
    '''
    This rule checks the **elsif** keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       ELSIF (a = '1') then

    **Fix**

    .. code-block:: vhdl

       elsif (a = '1') then
    '''

    def __init__(self):
        token_case.__init__(self, 'if', '026', lTokens)
        self.groups.append('case::keyword')




from vsg import parser


class when_keyword(parser.keyword):
    '''
    unique_id = conditional_expressions : when_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class else_keyword(parser.keyword):
    '''
    unique_id = conditional_expressions : else_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)





from vsg.rule_group import length
from vsg import violation


class line_length(length.Rule):
    '''
    Checks for a at least a single space before a token.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    '''

    def __init__(self, name, identifier):
        length.Rule.__init__(self, name=name, identifier=identifier)
        self.disable = False
        self.length = 120
        self.solution = 'Reduce line to less than ' + str(self.length) + ' characters'

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_lines_with_length_that_exceed_column(self.length)

    def _analyze(self, lToi):
        for oToi in lToi:
            oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
            self.add_violation(oViolation)





from vsg.token import interface_signal_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import expression
from vsg.vhdlFile.classify import identifier_list
from vsg.vhdlFile.classify import mode
from vsg.vhdlFile.classify import subtype_indication


def detect(iToken, lObjects):
    '''
    interface_signal_declaration ::=
        [ signal ] identifier_list : [ mode ] subtype_indication [ bus ] [ := *static*_expression ]
    '''

    if utils.is_next_token('signal', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('signal', token.signal_keyword, iToken, lObjects)

    iCurrent = identifier_list.classify_until([':'], iCurrent, lObjects, token.identifier)

    iCurrent = utils.assign_next_token_required(':', token.colon, iCurrent, lObjects)

    iCurrent = mode.classify(iCurrent, lObjects)

    iCurrent = subtype_indication.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_if('bus', token.bus_keyword, iCurrent, lObjects)

    if utils.is_next_token(':=', iCurrent, lObjects):

        iCurrent = utils.assign_next_token_required(':=', token.assignment, iCurrent, lObjects)

        iCurrent = expression.classify_until([')', ';'], iCurrent, lObjects)

    return iCurrent




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import entity_declaration as token


class rule_011(Rule):
    '''
    This rule checks for a single space after the **end** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end    entity fifo;

    **Fix**

    .. code-block:: vhdl

       end entity fifo;
    '''
    def __init__(self):
        Rule.__init__(self, 'entity', '011')
        self.left_token = token.end_keyword
        self.right_token = token.end_entity_keyword



def print_output(dRunInfo):
    '''
    Displays results to stdout in a compact format.

    Parameters:
      dRunInfo (dictionary)
    '''
    sOutputString = 'File: '
    sOutputString += dRunInfo['filename']

    sOutputString += ' '

    if dRunInfo['severities']['Error'] == 0:
        sOutputString += 'OK'
    else:
        sOutputString += 'ERROR'

    sOutputString += ' '

    sOutputString += '('
    sOutputString += str(dRunInfo['num_rules_checked'])
    sOutputString += ' rules checked)'

    for sSeverity in list(dRunInfo['severities'].keys()):
        sOutputString += ' ['
        sOutputString += sSeverity
        sOutputString += ': '
        sOutputString += str(dRunInfo['severities'][sSeverity])
        sOutputString += ']'
    if dRunInfo['severities']['Error'] == 0:
        return sOutputString, None
    else:
        return None, sOutputString




from vsg.token import concurrent_signal_assignment_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import concurrent_conditional_signal_assignment
from vsg.vhdlFile.classify import concurrent_selected_signal_assignment
from vsg.vhdlFile.classify import concurrent_simple_signal_assignment


def detect(iToken, lObjects):
    '''
    concurrent_signal_assignment_statement ::=
        [ label : ] [ postponed ] concurrent_simple_signal_assignment
      | [ label : ] [ postponed ] concurrent_conditional_signal_assignment
      | [ label : ] [ postponed ] concurrent_selected_signal_assignment
    '''

    iCurrent = iToken
    if concurrent_selected_signal_assignment.detect(iToken, lObjects):
        iCurrent = utils.tokenize_label(iCurrent, lObjects, token.label_name, token.label_colon)
        iCurrent = utils.tokenize_postponed(iCurrent, lObjects, token.postponed_keyword)
        iCurrent = concurrent_selected_signal_assignment.classify(iCurrent, lObjects)

    elif concurrent_conditional_signal_assignment.detect(iToken, lObjects):
        iCurrent = utils.tokenize_label(iCurrent, lObjects, token.label_name, token.label_colon)
        iCurrent = utils.tokenize_postponed(iCurrent, lObjects, token.postponed_keyword)
        iCurrent = concurrent_conditional_signal_assignment.classify(iCurrent, lObjects)

    elif concurrent_simple_signal_assignment.detect(iToken, lObjects):
        iCurrent = utils.tokenize_label(iCurrent, lObjects, token.label_name, token.label_colon)
        iCurrent = utils.tokenize_postponed(iCurrent, lObjects, token.postponed_keyword)
        iCurrent = concurrent_simple_signal_assignment.classify(iCurrent, lObjects)

    return iCurrent




from vsg.rules import remove_tokens

from vsg.token import case_statement as token

lTokens = []
lTokens.append(token.end_case_label)


class rule_020(remove_tokens):
    '''
    This rule checks for labels after the **end case** keywords.
    The label should be removed.
    The preference is to have comments above the case statement.

    **Violation**

    .. code-block:: vhdl

          end case CASE_LABEL;
          end case;

    **Fix**

    .. code-block:: vhdl

          end case;
          end case;
    '''

    def __init__(self):
        remove_tokens.__init__(self, 'case', '020', lTokens)
        self.solution = 'Remove Label'




from vsg.token import index_subtype_definition as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import type_mark


def classify(iToken, lObjects):
    '''
        index_subtype_definition ::=
            type_mark range <>
    '''

    iCurrent = type_mark.classify(iToken, lObjects)
    iCurrent = utils.assign_next_token_required('range', token.range_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('<>', token.undefined_range, iCurrent, lObjects)
    return iCurrent




from vsg.token import component_instantiation_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import generic_map_aspect
from vsg.vhdlFile.classify import instantiated_unit
from vsg.vhdlFile.classify import port_map_aspect


def detect(iToken, lObjects):
    '''
    component_instantiation_statement ::=
        instantiation_label :
            instantiated_unit
                [ generic_map_aspect ]
                [ port_map_aspect ] ;
    '''
    iCurrent = utils.find_next_token(iToken, lObjects)
    iCurrent = utils.increment_token_count(iCurrent)
    iCurrent = utils.find_next_token(iCurrent, lObjects)
    if not utils.object_value_is(lObjects, iCurrent, ':'):
        return iToken
    iCurrent = utils.increment_token_count(iCurrent)
    if instantiated_unit.detect(iCurrent, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.instantiation_label, token.label_colon)

    iCurrent = instantiated_unit.classify(iCurrent, lObjects)

    iCurrent = generic_map_aspect.detect(iCurrent, lObjects)

    iCurrent = port_map_aspect.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure


class rule_002(structure.Rule):
    '''
    This rule checks the boolean expression is enclosed in ().

    .. NOTE:: There is a configuration option **parenthesis** which will either insert or remove the parenthesis.

    parenthesis set to 'insert' (Default)
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    **Violation**

    .. code-block:: vhdl

      if a = '1' then

    **Fix**

    .. code-block:: vhdl

      if (a = '1') then

    parenthesis set to 'remove'
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^

    **Violation**

    .. code-block:: vhdl

      if (a = '1') then

    **Fix**

    .. code-block:: vhdl

      if a = '1' then
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'if', '002')
        self.parenthesis = 'insert'
        self.configuration.append('parenthesis')
        self.subphase = 2

    def _get_tokens_of_interest(self, oFile):
        if self.parenthesis == 'insert':
            return oFile.get_if_statement_conditions()
        return oFile.get_if_statement_conditions(fRemoveWhitespace=False)

    def _analyze(self, lToi):
        for oToi in lToi:
            if insert_parenthesis(self.parenthesis):
                check_insert_parenthesis(self, oToi)
            else:
                check_remove_parenthesis(self, oToi)

    def _fix_violation(self, oViolation):
        if insert_parenthesis(self.parenthesis):
            add_enclosing_parens(oViolation)
        else:
            remove_enclosing_parens(oViolation)


def remove_enclosing_parens(oViolation):
    lTokens = oViolation.get_tokens()
    dAction = oViolation.get_action()

    lNewTokens = remove_open_paren(lTokens, dAction)
    iDelta = len(lTokens) - len(lNewTokens)
    lNewTokens = remove_close_paren(lNewTokens, dAction, iDelta)
    oViolation.set_tokens(lNewTokens)


def remove_close_paren(lTokens, dAction, iDelta):
    lReturn = []
    for iToken, oToken in enumerate(lTokens):
        if iToken + iDelta not in dAction['right_remove']:
            lReturn.append(oToken)
    lReturn.extend(dAction['right_insert'])
    return lReturn


def remove_open_paren(lTokens, dAction):
    lReturn = []
    lReturn.extend(dAction['left_insert'])
    for iToken, oToken in enumerate(lTokens):
        if iToken not in dAction['left_remove']:
            lReturn.append(oToken)
    return lReturn


def add_enclosing_parens(oViolation):
    lTokens = oViolation.get_tokens()
    rules_utils.insert_token(lTokens, 0, parser.open_parenthesis())
    lTokens.append(parser.close_parenthesis())
    oViolation.set_tokens(lTokens)


def check_insert_parenthesis(self, oToi):
    lTokens = oToi.get_tokens()
    if not enclosing_parens_found(lTokens):
        oViolation = create_insert_violation(oToi)
        self.add_violation(oViolation)


def create_insert_violation(oToi):
    sSolution = 'Enclose condition in ()\'s.'
    dAction = {}
    dAction['action'] = 'insert'
    oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
    oViolation.set_action(dAction)
    return oViolation


def check_remove_parenthesis(self, oToi):
    lTokens = oToi.get_tokens()
    if enclosing_parens_found(lTokens):
        oViolation = create_remove_violation(oToi, lTokens)
        self.add_violation(oViolation)


def create_remove_violation(oToi, lTokens):
        sSolution = 'Remove enclosing ()\'s'
        dAction = create_remove_action_dict(lTokens)

        oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
        oViolation.set_action(dAction)
        return oViolation


def create_remove_action_dict(lTokens):
        dAction = {}
        dAction['action'] = 'remove'
        analyze_open_paren_cases(lTokens, dAction)
        analyze_close_paren_cases(lTokens, dAction)
        return dAction


def enclosing_parens_found(lTokens):
    if condition_starts_with_parenthesis(lTokens) and condition_ends_with_parenthesis(lTokens):
        lParens = build_parenthesis_list(lTokens)
        lNewParens = remove_inner_parenthesis(lParens)

        if len(lNewParens) == 2 and parens_match(lParens, lNewParens):
            return True
    return False


def analyze_close_paren_cases(lTokens, dAction):

    iLength = len(lTokens)
    if space_close_paren_space(lTokens):
        dAction['right_remove'] = [iLength - 1, iLength - 2]
        dAction['right_insert'] = []
    elif space_close_paren(lTokens):
        dAction['right_remove'] = [iLength - 1]
        dAction['right_insert'] = []
    elif close_paren_space(lTokens):
        dAction['right_remove'] = [iLength - 2]
        dAction['right_insert'] = []
    else:
        dAction['right_remove'] = [iLength - 1]
        dAction['right_insert'] = [parser.whitespace(' ')]


def analyze_open_paren_cases(lTokens, dAction):
    if open_paren_space(lTokens):
        dAction['left_remove'] = [0]
        dAction['left_insert'] = []
    elif space_open_paren_space(lTokens):
        dAction['left_remove'] = [0, 1]
        dAction['left_insert'] = []
    elif space_open_paren(lTokens):
        dAction['left_remove'] = [1]
        dAction['left_insert'] = []
    else:
        dAction['left_remove'] = [0]
        dAction['left_insert'] = [parser.whitespace(' ')]


def open_paren_space(lTokens):
    if isinstance(lTokens[0], parser.open_parenthesis) and isinstance(lTokens[1], parser.whitespace):
        return True
    return False


def space_open_paren_space(lTokens):
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], parser.open_parenthesis) and isinstance(lTokens[2], parser.whitespace):
        return True
    return False


def space_open_paren(lTokens):
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], parser.open_parenthesis):
        return True
    return False


def insert_parenthesis(option):
    if option == 'insert':
        return True
    return False


def space_close_paren(lTokens):
    if isinstance(lTokens[-1], parser.close_parenthesis) and isinstance(lTokens[-2], parser.whitespace):
        return True
    return False


def space_close_paren_space(lTokens):
    if isinstance(lTokens[-1], parser.whitespace) and isinstance(lTokens[-2], parser.close_parenthesis) and isinstance(lTokens[-3], parser.whitespace):
        return True
    return False


def close_paren_space(lTokens):
    if isinstance(lTokens[-1], parser.whitespace) and isinstance(lTokens[-2], parser.close_parenthesis):
        return True
    return False


def build_parenthesis_list(lTokens):
    lParens = []
    for oToken in lTokens:
       if isinstance(oToken, parser.open_parenthesis):
           lParens.append(oToken)
       elif isinstance(oToken, parser.close_parenthesis):
           lParens.append(oToken)
    return lParens


def remove_inner_parenthesis(lParens):
    lNewParens = []
    bSkipCloseParen = False
    if len(lParens) <= 2:
        return lParens

    for iParen, oParen in enumerate(lParens[:-1]):
        if bSkipCloseParen:
           bSkipCloseParen = False
           continue
        if isinstance(oParen, parser.open_parenthesis) and isinstance(lParens[iParen + 1], parser.close_parenthesis):
            bSkipCloseParen = True
            continue
        lNewParens.append(oParen)
    if isinstance(lParens[-2], parser.close_parenthesis) and isinstance(lParens[-1], parser.close_parenthesis):
       lNewParens.append(lParens[-1])

    lReturnParens = remove_inner_parenthesis(lNewParens)
    return lReturnParens


def parens_match(lParens, lNewParens):
    if lParens[0] == lNewParens[0] and lParens[-1] == lNewParens[-1]:
        return True
    return False


def condition_starts_with_parenthesis(lTokens):
    if isinstance(lTokens[0], parser.open_parenthesis):
        return True
    try:
        if isinstance(lTokens[1], parser.open_parenthesis):
            return True
    except IndexError:
        return False
    return False


def condition_ends_with_parenthesis(lTokens):
    if isinstance(lTokens[-1], parser.close_parenthesis):
        return True
    if isinstance(lTokens[-2], parser.close_parenthesis):
        return True
    return False




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.full_type_declaration.identifier, token.full_type_declaration.is_keyword])


class rule_006(Rule):
    '''
    This rule checks for a single space before the **is** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       type state_machine    is (idle, write, read, done);

    **Fix**

    .. code-block:: vhdl

       type state_machine is (idle, write, read, done);
    '''
    def __init__(self):
        Rule.__init__(self, 'type', '006', lTokens)




from vsg.rules import align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens

from vsg import parser
from vsg import token

lAlign = []
lAlign.append(parser.comment)

lSkip = []
lSkip.append(parser.comment)

oStart = token.package_declaration.is_keyword
oEnd = token.package_declaration.end_keyword


class rule_401(align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens):
    '''
    This rule checks the alignment of inline comments in the package declarative part.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       package my_package is

         signal   wr_en    : std_logic;  -- Comment 1
         signal   rd_en    : std_logic;     -- Comment 2
         constant c_period : time; -- Comment 3

       end package my_package;

    **Fix**

    .. code-block:: vhdl

       package my_package is

         signal   wr_en    : std_logic; -- Comment 1
         signal   rd_en    : std_logic; -- Comment 2
         constant c_period : time;      -- Comment 3

       end package my_package;
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens.__init__(self, 'package', '401', lAlign, oStart, oEnd, lSkip)
        self.solution = 'Align comments.'
        self.subphase = 4




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.exponent.e_keyword)


class rule_500(token_case):
    '''
    This rule checks the e keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

        12.5E-90
        6E57

    **Fix**

    .. code-block:: vhdl

        12.5e-90
        6e57
    '''

    def __init__(self):
        token_case.__init__(self, 'exponent', '500', lTokens)
        self.groups.append('case::keyword')




import bisect

from vsg.vhdlFile.extract import tokens


def get_tokens_between_non_whitespace_token_and_token(right_token, lAllTokens, oTokenMap):

    lReturn = []

    lStart, lEnd = get_start_and_end_indexes(right_token, lAllTokens, oTokenMap)

    for iStart, iEnd in zip(lStart, lEnd):
        if tokens_are_next_to_each_other(iStart, iEnd):
            continue

        iLine = oTokenMap.get_line_number_of_index(iStart)

        oTokens = tokens.New(iStart, iLine, lAllTokens[iStart:iEnd + 1])
        lReturn.append(oTokens)

    return lReturn


def get_start_and_end_indexes(right_token, lAllTokens, oTokenMap):
    lEnd = oTokenMap.get_token_indexes(right_token)
    lStart = []
    for iEnd in lEnd:
        lStart.append(oTokenMap.get_index_of_previous_non_whitespace_token(iEnd))
    return lStart, lEnd


def tokens_are_next_to_each_other(iStart, iEnd):
    if iStart == iEnd - 1:
        return True
    return False




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.instantiated_unit.entity_keyword)


class rule_027(token_case):
    '''
    This rule checks the **entity** keyword has proper case in direct instantiations.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       INSTANCE_NAME : ENTITY library.ENTITY_NAME

    **Fix**

    .. code-block:: vhdl

       INSTANCE_NAME : entity library.ENTITY_NAME
    '''

    def __init__(self):
        token_case.__init__(self, 'instantiation', '027', lTokens)
        self.groups.append('case::keyword')




from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import entity_statement


def detect(iToken, lObjects):
    '''
    entity_statement_part ::=
        { entity_statement }
    '''

    return utils.detect_submodule(iToken, lObjects, entity_statement)




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens(structure.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    token : token object type
       The token to insert a carriage return after.

    oStart : token object type
       The starting token which defines the range

    oEnd : token object type
       The ending token which defines the range
    '''

    def __init__(self, name, identifier, lTokens, oStart, oEnd):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens
        self.oStart = oStart
        self.oEnd = oEnd

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_token_and_n_tokens_after_it_when_between_tokens(self.lTokens, 2, self.oStart, self.oEnd)

    def _analyze(self, lToi):
        for oToi in lToi:
           lTokens = oToi.get_tokens()
           if utils.are_next_consecutive_token_types([parser.carriage_return], 1, lTokens):
               continue
           if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment], 1, lTokens):
               continue
           if utils.are_next_consecutive_token_types([parser.comment, parser.carriage_return], 1, lTokens):
               continue
           if utils.are_next_consecutive_token_types([parser.whitespace, parser.carriage_return], 1, lTokens):
               continue
           else:
               self.add_violation(violation.New(oToi.get_line_number(), oToi, self.solution))

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        rules_utils.insert_carriage_return(lTokens, 1)
        oViolation.set_tokens(lTokens)




from vsg.token import interface_procedure_specification as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import formal_parameter_list


def detect(iToken, lObjects):
    '''
    interface_procedure_specification ::=
        procedure designator
            [ [ parameter ] ( formal_parameter_list ) ]
    '''
    if utils.is_next_token('procedure', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('procedure', token.procedure_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token(token.designator, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if('parameter', token.parameter_keyword, iCurrent, lObjects)

    if utils.is_next_token('(', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iCurrent, lObjects)
        iCurrent = formal_parameter_list.classify(iCurrent, lObjects)
        iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)

    return iCurrent




from vsg import parser


class mode(parser.keyword):
    '''
    unique_id = mode : mode
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class in_keyword(mode):
    '''
    unique_id = mode : in_keyword
    '''

    def __init__(self, sString):
        mode.__init__(self, sString)


class out_keyword(mode):
    '''
    unique_id = mode : out_keyword
    '''

    def __init__(self, sString):
        mode.__init__(self, sString)


class inout_keyword(mode):
    '''
    unique_id = mode : inout_keyword
    '''

    def __init__(self, sString):
        mode.__init__(self, sString)


class buffer_keyword(mode):
    '''
    unique_id = mode : buffer_keyword
    '''

    def __init__(self, sString):
        mode.__init__(self, sString)


class linkage_keyword(mode):
    '''
    unique_id = mode : linkage_keyword
    '''

    def __init__(self, sString):
        mode.__init__(self, sString)




from vsg import deprecated_rule


class rule_002(deprecated_rule.Rule):
    '''
    This rule has been superceeded by:

    * `attribute_declaration_500 <attribute_declaration_rules.html#attribute-declaration-500>`_
    * `attribute_specification_500 <attribute_specification_rules.html#attribute-specification-500>`_
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'attribute', '002')
        self.message.append('Rule ' + self.unique_id + ' has been superceeded by attribute_declaration_500 and attribute_specification_500.')




from vsg import parser


class open_keyword(parser.keyword):
    '''
    unique_id = file_open_information : open_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class file_open_kind_expression(parser.expression):
    '''
    unique_id = file_open_information : file_open_kind_expression
    '''

    def __init__(self, sString):
        parser.expression.__init__(self, sString)


class is_keyword(parser.keyword):
    '''
    unique_id = file_open_information : is_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class file_logical_name(parser.expression):
    '''
    unique_id = file_open_information : file_logical_name
    '''

    def __init__(self, sString):
        parser.expression.__init__(self, sString)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.library_clause.keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of the **library** keyword.
    Indenting helps in comprehending the code.


    **Violation**

    .. code-block:: vhdl

       library ieee;
          library fifo_dsn;

    **Fix**

    .. code-block:: vhdl

       library ieee;
       library fifo_dsn;
    '''

    def __init__(self):
        token_indent.__init__(self, 'library', '001', lTokens)




from vsg.rules import previous_line

from vsg import token

lTokens = []
lTokens.append(token.loop_statement.loop_label)
lTokens.append(token.iteration_scheme.while_keyword)
lTokens.append(token.iteration_scheme.for_keyword)
lTokens.append(token.loop_statement.loop_keyword)


class rule_200(previous_line):
    '''
    This rule checks for blank lines or comments above loop statements.

    |configuring_blank_lines_link|

    The default style is :code:`no_code`.

    **Violation**

    .. code-block:: vhdl

       -- Comment
       LOOP_LABEL : loop

       wr_en <= wr_en;
       LOOP_LABEL : loop

    **Fix**

    .. code-block:: vhdl

       -- Comment
       LOOP_LABEL : loop

       wr_en <= wr_en;

       LOOP_LABEL : loop
    '''

    def __init__(self):
        previous_line.__init__(self, 'loop_statement', '200', lTokens)
        self.style = 'no_code'




from vsg.token import group_constituent_list as token

from vsg.vhdlFile import utils


def classify(iToken, lObjects):
    '''
    association_list ::=
        association_element { , association_element }
    '''
    iCurrent = utils.assign_next_token(token.group_constituent, iToken, lObjects)

    while utils.is_next_token(',', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required(',', token.comma, iCurrent, lObjects)
        iCurrent = utils.assign_next_token(token.group_constituent, iToken, lObjects)

    return iCurrent




from vsg import parser


class on_keyword(parser.keyword):
    '''
    unique_id = sensitivity_clause : on_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import loop_statement as token


class rule_100(Rule):
    '''
    This rule checks that a single space exists between the **end** and **loop** keywords

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

         end loop;
         end    loop;

    **Fix**

    .. code-block:: vhdl

         end loop;
         end loop;
    '''
    def __init__(self):
        Rule.__init__(self, 'loop_statement', '100')
        self.left_token = token.end_keyword
        self.right_token = token.end_loop_keyword




from vsg.rules import token_case_in_range_bounded_by_tokens as Rule

from vsg import token

lTokens = []
lTokens.append(token.choice.others_keyword)

oStartToken = token.case_generate_alternative.when_keyword

oEndToken = token.case_generate_alternative.assignment


class rule_501(Rule):
    '''
    This rule checks the *others* keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       when OTHERS =>

    **Fix**

    .. code-block:: vhdl

       when others =>
    '''

    def __init__(self):
        Rule.__init__(self, 'case_generate_alternative', '501', lTokens, oStartToken, oEndToken)
        self.groups.append('case::keyword')




from vsg.rules import token_case as Rule

from vsg.token import predefined_attribute as token

lTokens = []
lTokens.append(token.keyword)


class rule_500(Rule):
    '''
    This rule checks predefined attributes have the proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       signal data : std_logic_vector(g_width'RANGE);

    **Fix**

    .. code-block:: vhdl

       signal data : std_logic_vector(g_width'range);
    '''

    def __init__(self):
        Rule.__init__(self, 'attribute', '500', lTokens)
        self.groups.append('case::keyword')




from vsg.rules import previous_line

from vsg.token import entity_declaration as token


class rule_003(previous_line):
    '''
    This rule checks for blank lines or comments above the entity keyword.

    |configuring_previous_line_rules_link|

    **Violation**

    .. code-block:: vhdl

       library ieee;
       entity fifo is

    **Fix**

    .. code-block:: vhdl

       library ieee;

       entity fifo is
    '''

    def __init__(self):
        previous_line.__init__(self, 'entity', '003', [token.entity_keyword])




from vsg import deprecated_rule


class rule_005(deprecated_rule.Rule):
    '''
    This rule has been deprecated and it's function has been included in rules **port_007**, **port_008** and **port_009**.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'port', '005')
        self.message.append('Rule ' + self.unique_id + ' has been included in rules port_007, port_008 and port_009.')




import os


def get_violation_line_number(dViolation):
    '''
    Returns a line number of a violation.

    Parameters:

      dViolation: Violation dictionary

    Returns:  integer
    '''
    return dViolation.get_line_number()


def get_violation_at_line_number(lViolations, iLineNumber):
    '''
    Returns a violation from a violation dictionary at the given line number.

    Parameters:

      lViolation : (List of Violation dictionaries)

      iViolation : (integer)

    Return: Violation Dictionary
    '''
    for oViolation in lViolations:
        if oViolation.get_line_number() == iLineNumber:
            return oViolation


def get_violation_solution_at_line_number(lViolations, iLineNumber):
    '''
    Returns a the solution for a violation at a given line number.

    Parameters:

      lViolation : (List of Violation dictionaries)

      iViolation : (integer)

    Return: string
    '''
    try:
        dViolation = get_violation_at_line_number(lViolations, iLineNumber)
        return dViolation['solution']
    except TypeError:
        oViolation = get_violation_at_line_number(lViolations, iLineNumber)
        try:
            return oViolation.get_solution()
        except AttributeError:
            return None


def expand_filename(sFileName):
    '''Expands environment variables in filenames.'''
    return os.path.expanduser(os.path.expandvars(sFileName))


def write_junit_xml_file(oJunitFile):
    with open(oJunitFile.filename, 'w') as oFile:
        for sLine in oJunitFile.build_junit():
            oFile.write(sLine + '\n')


def extract_file_names_from_file_list(lFiles):
    lReturn = []
    for dFile in lFiles:
        if isinstance(dFile, dict):
            lReturn.extend(extract_keys_from_dict(dFile))
        else:
            lReturn.append(dFile)

    return lReturn


def extract_keys_from_dict(dFile):
    lReturn = []
    lReturn.extend(list(dFile.keys()))
    return lReturn




from vsg.rules import token_suffix as Rule

from vsg import token

lTokens = []
lTokens.append(token.alias_declaration.alias_designator)


class rule_601(Rule):
    '''
    This rule checks for valid suffixes on alias designators.

    Default prefix is *\_a*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       alias header is name;
       alias footer is name;

    **Fix**

    .. code-block:: vhdl

       alias header_a is name;
       alias footer_a is name;
    '''

    def __init__(self):
        Rule.__init__(self, 'alias_declaration', '601', lTokens)
        self.suffixes = ['_a']
        self.solution = 'Alias designators'




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.block_statement.block_label)


class rule_300(token_indent):
    '''
    This rule checks the indent of the block label.

    **Violation**

    .. code-block:: vhdl

       a <= b;

          block_label : block is

    **Fix**

    .. code-block:: vhdl

       a <= b;

       block_label : block is
    '''

    def __init__(self):
        token_indent.__init__(self, 'block', '300', lTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.variable_assignment_statement.label)
lTokens.append(token.simple_variable_assignment.target)
lTokens.append(token.conditional_variable_assignment.target)
lTokens.append(token.selected_variable_assignment.with_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of a variable assignment.

    **Violation**

    .. code-block:: vhdl

       proc : process () is
       begin

           counter := 0;
       count := counter + 1;


    **Fix**

    .. code-block:: vhdl

       proc : process () is
       begin

         counter := 0;
         count   := counter + 1;

    '''

    def __init__(self):
        token_indent.__init__(self, 'variable_assignment', '001', lTokens)




from vsg.deprecated_rule import Rule


class rule_005(Rule):
    '''
    This rule has been deprecated and replaced with rule `process_400 <process_rules.html#process-400>`_.
    '''

    def __init__(self):
        Rule.__init__(self, 'variable_assignment', '005')
        self.message.append('Rule ' + self.unique_id + ' has been merged into process_400.')




from vsg import violation
from vsg.rule_group import structure


class rule_001(structure.Rule):
    '''
    This rule checks for the existance of the source file passed to VSG.

    **Violation**

    Source file passed to VSG does not exist.
    This violation will be reported at the command line in the normal output.
    It will also be reported in the junit file if the --junit option is used.

    **Fix**

    Pass correct file name to VSG.
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'source_file', '001')
        # These are filled out when creating a new rule
        self.fixable = False
        self.solution = 'File empty.'

    def analyze(self, oFile):
        oToi = oFile.get_all_tokens()
        iStartIndex = oToi.get_start_index()
        iEndIndex = oToi.get_end_index()

        if oFile.eError:
            oViolation = violation.New(0, oToi, oFile.eError.args[1])
            self.add_violation(oViolation)

        elif iStartIndex == iEndIndex:
            oViolation = violation.New(0, oToi, self.solution)
            self.add_violation(oViolation)




from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import concurrent_statement


def classify_until(lUntils, iToken, lObjects):
    '''
    architecture_statement_part ::=
        { concurrent_statement }
    '''

    return utils.detect_subelement_until(lUntils[0], concurrent_statement, iToken, lObjects)




from vsg import parser
from vsg import token
from vsg import violation

from vsg.rules import alignment_utils
from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils

lTokens = []
lTokens.append(parser.comment)

lSkip = []
lSkip.append(parser.comment)

oLeftToken = token.architecture_body.begin_keyword

oRightToken = token.architecture_body.end_keyword

lStart = []
lStart.append(token.concurrent_selected_signal_assignment.target)
lStart.append(token.concurrent_conditional_signal_assignment.target)
lStart.append(token.concurrent_simple_signal_assignment.target)

lEnd = []
lEnd.append(token.concurrent_selected_signal_assignment.semicolon)
lEnd.append(token.concurrent_conditional_signal_assignment.semicolon)
lEnd.append(token.concurrent_simple_signal_assignment.semicolon)


class rule_008(alignment.Rule):
    '''
    This rule checks the alignment of inline comments in consecutive concurrent statements.
    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <= '0';     -- Write enable
       rd_en <= '1';   -- Read enable
       data  <= (others => '0'); -- Write data

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0';             -- Write enable
       rd_en <= '1';             -- Read enable
       data  <= (others => '0'); -- Write data
    '''

    def __init__(self):
        alignment.Rule.__init__(self, 'concurrent', '008')
        self.subphase = 3
        self.lTokens = lTokens
        self.left_token = oLeftToken
        self.right_token = oRightToken
        self.lSkip = lSkip
        self.lStart = lStart
        self.lEnd = lEnd
        ## Stuff below is from original keyword_alignment_rule
        self.compact_alignment = True
        self.configuration.append('compact_alignment')

        self.blank_line_ends_group = True
        self.configuration.append('blank_line_ends_group')
        self.comment_line_ends_group = True
        self.configuration.append('comment_line_ends_group')
        self.include_lines_without_comments = False
        self.configuration.append('include_lines_without_comments')

    def analyze(self, oFile):

        lToi = oFile.get_tokens_bounded_by(self.left_token, self.right_token)
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
            dAnalysis = {}
            iColumn = 0
            iToken = -1
            iLeftColumn = 0
            bStartFound = False
            bEndFound = False

            for iIndex in range(0, len(lTokens)):
               iToken += 1
               oToken = lTokens[iIndex]
#               print(f'{oToken} | {oToken.get_value()}')

               if bStartFound:
                   if isinstance(oToken, parser.carriage_return):
                       if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment], iIndex + 1, lTokens) or \
                          utils.are_next_consecutive_token_types([parser.comment], iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)
                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)
                           dAnalysis = {}
                           bStartFound = False
                           bEndFound = False

                           iLine += 1
                           iLeftColumn = 0
                           iColumn = 0
                           iToken = -1
                           continue

                   elif isinstance(oToken, parser.comment):
#                       print(f'--> Comment Found     | {iLine} | {iColumn}')
                       dAnalysis[iLine] = {}
                       dAnalysis[iLine]['token_column'] = iColumn
                       dAnalysis[iLine]['token_index'] = iToken
                       dAnalysis[iLine]['line_number'] = iLine
                       if isinstance(lTokens[iIndex -1], parser.whitespace):
                           dAnalysis[iLine]['left_column'] = iColumn - len(lTokens[iIndex - 1].get_value())
                       else:
                           dAnalysis[iLine]['left_column'] = iColumn

               if isinstance(oToken, parser.carriage_return):
                   iLine += 1
                   iLeftColumn = 0
                   iColumn = 0
                   iToken = -1
               else:
                   iColumn += alignment_utils.update_column_width(self, oToken)
                   iLeftColumn += alignment_utils.update_column_width(self, oToken)

               if bEndFound:
                   for oStartToken in self.lStart:
                       if isinstance(oToken, oStartToken):
                           bStartFound = False
                           bEndFound = False
                           break
                   else:
                       if not isinstance(oToken, parser.whitespace) and \
                          not isinstance(oToken, parser.comment) and \
                          not isinstance(oToken, parser.carriage_return):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)
                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)
                           dAnalysis = {}
                           bStartFound = False
                           bEndFound = False
                           continue

               if bStartFound:
                   for oEndToken in self.lEnd:
                       if isinstance(oToken, oEndToken):
#                           print(f'--> End Token Found   | {iLine} | {iColumn}')
                           bEndFound = True
                           break
               else:
                   for oStartToken in self.lStart:
                       if isinstance(oToken, oStartToken):
#                           print(f'--> Start Token Found | {iLine} | {iColumn} | {oToken}')
                           bStartFound = True
                           break

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        iTokenIndex = dAction['token_index']

        if isinstance(lTokens[iTokenIndex - 1], parser.whitespace):
            iLen = len(lTokens[iTokenIndex - 1].get_value())
            lTokens[iTokenIndex - 1].set_value(' '*(iLen + dAction['adjust']))
        else:
            rules_utils.insert_whitespace(lTokens, iTokenIndex, dAction['adjust'])
        oViolation.set_tokens(lTokens)


def add_adjustments_to_dAnalysis(dAnalysis, compact_alignment, include_lines_without_comments=False, iMaxColumn=0):
    iMaxLeftColumn = 0
    iMaxTokenColumn = 0
    for iKey in list(dAnalysis.keys()):
        iMaxLeftColumn = max(iMaxLeftColumn, dAnalysis[iKey]['left_column'])
        iMaxTokenColumn = max(iMaxTokenColumn, dAnalysis[iKey]['token_column'])

    if include_lines_without_comments:
        iMaxTokenColumn = max(iMaxTokenColumn, iMaxColumn)

    if compact_alignment:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxLeftColumn - dAnalysis[iKey]['token_column'] + 1
    else:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxTokenColumn - dAnalysis[iKey]['token_column']




from vsg.token import next_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import condition


def detect(iToken, lObjects):
    '''
    next_statement ::=
        [ label : ] next [ loop_label ] [ when condition ] ;
    '''
    if utils.are_next_consecutive_tokens([None, ':', 'next'], iToken, lObjects):
        return classify(iToken, lObjects)
    if utils.is_next_token('next', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.label, token.label_colon)
    iCurrent = utils.assign_next_token_required('next', token.next_keyword, iCurrent, lObjects)

    if not utils.is_next_token(';', iCurrent, lObjects) and not utils.is_next_token('when', iCurrent, lObjects):
        iCurrent = utils.assign_next_token(token.loop_label, iCurrent, lObjects)

    if utils.is_next_token('when', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('when', token.when_keyword, iCurrent, lObjects)
        iCurrent = condition.classify_until([';'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




iLineLength = 80


def print_output(dRunInfo):
    '''
    Displays results to stdout in standard VSG format.

    Parameters:
      dRunInfo (dictionary)
    '''
    header = print_header(dRunInfo['filename'])
    stats = print_stats(dRunInfo)
    violations = print_violations(dRunInfo)
    return header + stats + violations, None


def print_header(sFilename):
    '''
    Prints the header for displaying results.

    Parameters:
      sFilename (string)

    Returns: Nothing
    '''
    top = '=' * iLineLength + '\n'
    filename = 'File:  ' + sFilename + '\n'
    bottom = '=' * iLineLength + '\n'
    return top + filename + bottom


def print_phase_information(iPhaseNumber):
    '''
    Prints information about the analysis.
    '''
    return 'Phase ' + str(iPhaseNumber) + ' of 7... Reporting\n'


def print_divider():
    '''
    Prints a divider that matches the column divisions in the violations.
    '''
    return '-'*28 + '+' + '-'*12 + '+' + '-'*12 + '+' + '-'*38 + '\n'


def print_violation_header():
    top_divider = print_divider()
    sOutputString = '  '
    sOutputString += 'Rule'.ljust(25)
    sOutputString += ' | '
    sOutputString += 'severity'.center(10)
    sOutputString += ' | '
    sOutputString += 'line(s)'.center(10)
    sOutputString += ' | '
    sOutputString += 'Solution\n'
    bottom_divider = print_divider()
    return top_divider + sOutputString + bottom_divider


def print_stats(dRunInfo):
    '''
    Displays information about the run such as number failures.

    Parameters:

      dRunInfo : (dictionary)

    Returns: Nothing
    '''
    stats = 'Phase ' + str(dRunInfo['stopPhase']) + ' of 7... Reporting\n'
    stats += 'Total Rules Checked: ' + str(dRunInfo['num_rules_checked']) + '\n'
    stats += 'Total Violations:    ' + str(dRunInfo['total_violations']) + '\n'
    for sSeverity in list(dRunInfo['severities'].keys()):
        sFormat = '  {0:<' + str(dRunInfo['maxSeverityNameLength']) + 's} : {1:5d}'
        stats += sFormat.format(sSeverity, dRunInfo['severities'][sSeverity]) + '\n'
    return stats


def print_violations(dRunInfo):
    '''
    Displays information about each violations.

    Parameters:

      dRunInfo : (dictionary)

    Returns: Nothing
    '''
    if dRunInfo['total_violations'] > 0:
        sOutputString = print_violation_header()
        for dViolation in dRunInfo['violations']:
            sOutputString += '  '
            sOutputString += dViolation['rule'].ljust(25)
            sOutputString += ' | '
            sOutputString += dViolation['severity']['name'].ljust(10)
            sOutputString += ' | '
            sOutputString += dViolation['lineNumber'].rjust(10)
            sOutputString += ' | '
            try:
                sOutputString += dViolation['solution']
            except TypeError:
                sOutputString += 'None'
            sOutputString += '\n'
        sOutputString += print_divider()
        sOutputString += 'NOTE: Refer to online documentation at https://vhdl-style-guide.readthedocs.io/en/latest/index.html for more information.'
        return sOutputString
    else:
        return ''




from vsg.token import primary_unit_declaration as token

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    primary_unit_declaration ::= identifier;
    '''
    return classify(iToken, lObjects)


def classify(iToken, lObjects):
    iCurrent = iToken

    while not utils.is_next_token(';', iCurrent, lObjects):
        iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_token(lObjects, iCurrent, token.semicolon)

    return iCurrent




from vsg.rules import token_indent as Rule


class token_indent_between_tokens(Rule):
    '''
    Checks the indent of tokens.
    '''

    def __init__(self, name, identifier, lTokens, oStart, oEnd, bInclusive=False):
        Rule.__init__(self, name=name, identifier=identifier, lTokens=lTokens)
        self.oStart = oStart
        self.oEnd = oEnd
        self.bInclusive = bInclusive

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_at_beginning_of_line_matching_between_tokens(self.lTokens, self.oStart, self.oEnd, self.bInclusive)




from vsg.token import loop_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import iteration_scheme
from vsg.vhdlFile.classify import sequence_of_statements


def detect(iToken, lObjects):
    '''
    loop_statement ::=
        [ loop_label : ]
            [ iteration_scheme ] loop
                sequence_of_statements
            end loop [ loop_label ] ;
    '''
    if utils.find_in_next_n_tokens(':', 2, iToken, lObjects):
        iCurrent = utils.find_next_token(iToken, lObjects)
        iCurrent += 1
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        iCurrent += 1
    else:
        iCurrent = iToken

    if iteration_scheme.detect(iCurrent, lObjects):
        return classify(iToken, lObjects)
    if utils.is_next_token('loop', iCurrent, lObjects):
        return classify(iToken, lObjects)

    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.loop_label, token.label_colon)

    iCurrent = iteration_scheme.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('loop', token.loop_keyword, iCurrent, lObjects)

    iCurrent = sequence_of_statements.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('loop', token.end_loop_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.end_loop_label, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.token import choices as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import choice


def classify_until(lUntils, iToken, lObjects):
    '''
    choices ::=
        choice { | choice }
    '''
    iCurrent = iToken
    iLast = 0
    lMyUntils = lUntils
    lMyUntils.append('|')

    while iLast != iCurrent:
        iLast = iCurrent
        iCurrent = choice.classify_until(lMyUntils, iCurrent, lObjects)
        iCurrent = utils.assign_next_token_if('|', token.bar, iCurrent, lObjects)

    return iCurrent




from vsg import parser

from vsg.vhdlFile import utils


def get_tokens_bounded_by(lTokenPairs, oFile):
    lToi = []
    for lTokenPair in lTokenPairs:
        aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
        lToi = utils.combine_two_token_class_lists(lToi, aToi)
    return lToi


def adjust_start_index_based_on_whitespace(oToi, iAdjustment):
    iStart = oToi.get_meta_data('iStart')
    lTokens = oToi.get_tokens()
    if isinstance(lTokens[iStart - 1], parser.whitespace):
        oToi.set_meta_data('iStart', iStart + iAdjustment)




from vsg.rules import previous_line as Rule

from vsg.token import case_statement_alternative as token


class rule_201(Rule):
    '''
    This rule checks for blank lines or comments above the **when** keyword.

    Refer to `Configuring Previous Line Rules <configuring.html#configuring-previous-line-rules>`_ for options.

    The default style is :code:`allow_comment`.

    **Violation**

    .. code-block:: vhdl

       case data is
         when 3 =>
         -- Comment
         when 4 =>

    **Fix**

    .. code-block:: vhdl

       case data is

         when 3 =>

         -- Comment
         when 4 =>
    '''

    def __init__(self):
        Rule.__init__(self, 'case', '201', [token.when_keyword])
        self.style = 'allow_comment'




from vsg.rules import insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token

from vsg.token import block_statement as token

oInsertToken = token.end_block_label

oAnchorToken = token.semicolon

oLeftToken = token.end_keyword
oRightToken = token.semicolon

oValueToken = token.block_label


class rule_007(insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token):
    '''
    This rule checks the block label exists in the closing of the block statement.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       end block;

    **Fix**

    .. code-block:: vhdl

       end block block_label;
    '''

    def __init__(self):
        insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token.__init__(self, 'block', '007', oInsertToken, oAnchorToken, oLeftToken, oRightToken, oValueToken)
        self.solution = 'label'
        self.groups.append('structure::optional')




from vsg import parser
from vsg import violation

from vsg.rule_group import whitespace


class remove_spaces_before_token_rule(whitespace.Rule):
    '''
    This class removes whitespace before a given token.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    oToken : token object
       The token where spaces will be removed before.
    '''

    def __init__(self, name, identifier, oToken, bIgnoreIfLineStart=False):

        whitespace.Rule.__init__(self, name, identifier)
        self.oToken = oToken
        self.bIgnoreIfLineStart = bIgnoreIfLineStart

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_sequence_of_tokens_matching([parser.whitespace, self.oToken], self.bIgnoreIfLineStart)

    def _analyze(self, lToi):
        for oToi in lToi:
            self.add_violation(violation.New(oToi.get_line_number(), oToi, self.solution))

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        oViolation.set_tokens(lTokens[1:])




from vsg import exceptions
from vsg import parser

from vsg.token import direction
from vsg.token import choice
from vsg.token import element_association
from vsg.token import exponent
from vsg.token import predefined_attribute

from vsg.token.ieee.std_logic_1164 import types


def assign_tokens_until(sToken, token, iToken, lObjects):
    iCurrent = iToken
    while not is_next_token(sToken, iCurrent, lObjects):
        iCurrent = assign_next_token(token, iCurrent, lObjects)
    return iCurrent


def assign_next_token(token, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    try:
        lObjects[iCurrent] = token(lObjects[iCurrent].get_value())
    except TypeError:
        lObjects[iCurrent] = token()
    iCurrent+= 1
    return iCurrent


def assign_token(lObjects, iToken, token):
    iCurrent = find_next_token(iToken, lObjects)
    try:
        lObjects[iCurrent] = token(lObjects[iCurrent].get_value())
    except TypeError:
        lObjects[iToken] = token()
    return iToken + 1


def assign_next_token_if(sToken, token, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    if object_value_is(lObjects, iCurrent, sToken):
        lObjects[iCurrent] = token(lObjects[iCurrent].get_value())
        iCurrent += 1
        return iCurrent
    return iToken


def assign_next_token_if_not(sToken, token, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    if not object_value_is(lObjects, iCurrent, sToken):
        lObjects[iCurrent] = token(lObjects[iCurrent].get_value())
        iCurrent += 1
        return iCurrent
    return iToken


def assign_next_token_if_not_one_of(lTokens, token, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    if lObjects[iCurrent].get_value().lower() not in lTokens:
        lObjects[iCurrent] = token(lObjects[iCurrent].get_value())
        iCurrent += 1
        return iCurrent
    return iToken


def assign_next_token_required(sToken, token, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    if object_value_is(lObjects, iCurrent, sToken):
        lObjects[iCurrent] = token(lObjects[iCurrent].get_value())
        return iCurrent + 1
    else:
        print_error_message(sToken, token, iCurrent, lObjects)
    return iToken


def assign_tokens_until_matching_closing_paren(token, iToken, lObjects):
    iCounter = 1
    iCurrent = iToken
    while iCurrent < len(lObjects):
        iCurrent = find_next_token(iCurrent, lObjects)
        iCounter = update_paren_counter(iCurrent, lObjects, iCounter)
        if token_is_close_parenthesis(iCurrent, lObjects) and iCounter == 0:
            return iCurrent
        lObjects[iCurrent] = token(lObjects[iCurrent].get_value())


def object_value_is(lAllObjects, iToken, sString):
    if lAllObjects[iToken].get_value().lower() == sString.lower():
        return True
    return False


def is_item(lAllObjects, iToken):
    if type(lAllObjects[iToken]) == parser.item:
        return True
    return False


def get_range(lObjects, iStart, sEnd):
    iIndex = iStart
    while lObjects[iIndex].get_value().lower() != sEnd:
        iIndex += 1
    iEnd = iIndex
    iStart = iStart
    return iStart, iEnd


def classify_next_token(token, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    iCurrent = assign_token(lObjects, iCurrent, token)
    return iCurrent


def find_in_range(sValue, iToken, sEnd, lObjects):
    iStart, iEnd = get_range(lObjects, iToken, sEnd)
    for iIndex in range(iStart, iEnd + 1):
        if object_value_is(lObjects, iIndex, sValue):
            return True
    return False


def find_in_index_range(sValue, iStart, iEnd, lObjects):
    for iIndex in range(iStart, iEnd + 1):
        if object_value_is(lObjects, iIndex, sValue):
            return True
    return False


def are_next_consecutive_tokens(lTokens, iToken, lObjects):
    iMaxTokenCount = len(lTokens)
    iTokenCount = 0
    iCurrent = iToken
    while iTokenCount < iMaxTokenCount:
        iCurrent = find_next_token(iCurrent, lObjects)
        if not lTokens[iTokenCount] is None:
            if not is_next_token(lTokens[iTokenCount], iCurrent, lObjects):
                return False
        iCurrent += 1
        iTokenCount += 1

    return True


def are_next_consecutive_token_types(lTypes, iToken, lObjects):
    iMaxTokenCount = len(lTypes)
    iTokenCount = 0
    iCurrent = iToken
    try:
        while iTokenCount < iMaxTokenCount:
            if not lTypes[iTokenCount] is None:
                if not isinstance(lObjects[iCurrent], lTypes[iTokenCount]):
                    return False
            iCurrent += 1
            iTokenCount += 1
        return True
    except IndexError:
        return False


def are_next_consecutive_token_types_ignoring_whitespace(lTypes, iToken, lObjects):
    iMaxTokenCount = len(lTypes)
    iTokenCount = 0
    iCurrent = iToken
    try:
        while iTokenCount < iMaxTokenCount:
            iCurrent = find_next_non_whitespace_token(iCurrent, lObjects)
            if not lTypes[iTokenCount] is None:
                if not isinstance(lObjects[iCurrent], lTypes[iTokenCount]):
                    return False
            iCurrent += 1
            iTokenCount += 1
        return True
    except IndexError:
        return False

def are_previous_consecutive_token_types_ignoring_whitespace(lTypes, iToken, lObjects):
    iMinTokenCount = 0
    iTokenCount = len(lTypes)
    iCurrent = iToken
    try:
        while iTokenCount != iMinTokenCount:
            iCurrent = find_previous_non_whitespace_token(iCurrent, lObjects)
            if not lTypes[iTokenCount - 1] is None:
                if not isinstance(lObjects[iCurrent], lTypes[iTokenCount - 1]):
                    return False
            iCurrent -= 1
            iTokenCount -= 1
        return True
    except IndexError:
        return False


def find_in_next_n_tokens(sValue, iMax, iToken, lObjects):
    iEnd = len(lObjects)
    iTokenCount = 0
    iCurrent = iToken
    while iTokenCount < iMax:
        iCurrent = find_next_token(iCurrent, lObjects)
        iTokenCount += 1
        if object_value_is(lObjects, iCurrent, sValue):
            return True
        iCurrent += 1
        if iCurrent == iEnd:
            return False
    return False


def find_earliest_occurance(lEnd, iToken, lObjects):
    iEarliest = 9999999999999999999999999999
    for sEnd in lEnd:
        for iIndex in range(iToken, len(lObjects) - 1):
            if lObjects[iIndex].get_value().lower() == sEnd:
                if iIndex < iEarliest:
                    sEarliest = lObjects[iIndex].get_value()
                    iEarliest = iIndex
    return sEarliest


def find_next_token(iToken, lObjects):
    for iCurrent in range(iToken, len(lObjects)):
        if is_item(lObjects, iCurrent):
            return iCurrent
    return iToken


def find_next_non_whitespace_token(iToken, lObjects):
    iCurrent = iToken
    for iIndex in range(iToken, len(lObjects)):
        oToken = lObjects[iIndex]
        if token_is_whitespace_or_comment(oToken):
            continue
        return iIndex
    return iCurrent


def find_previous_non_whitespace_token(iToken, lObjects):
    iCurrent = iToken
    for iIndex in range(iToken, -1, -1):
        oToken = lObjects[iIndex]
        if token_is_whitespace_or_comment(oToken):
            continue
        return iIndex
    return iCurrent


def detect_submodule(iToken, lObjects, module):
    iLast = -1
    iReturn = iToken
    while iLast != iReturn:
        if is_next_token('end', iReturn, lObjects):
            return iToken
        iReturn = find_next_token(iReturn, lObjects)
        iLast = iReturn
        iReturn = module.detect(iReturn, lObjects)

    return iReturn


def has_label(iObject, lObjects):
    iCurrent = find_next_token(iObject, lObjects)
    iCurrent = increment_token_count(iCurrent)
    iCurrent = find_next_token(iCurrent, lObjects)
    if object_value_is(lObjects, iCurrent, ':'):
        return True
    return False


def tokenize_postponed(iObject, lObjects, token):
    iIndex = find_next_token(iObject, lObjects)
    if object_value_is(lObjects, iIndex, 'postponed'):
        assign_token(lObjects, iIndex, token)
        return iIndex + 1
    return iObject


def tokenize_label(iToken, lObjects, label_token, colon_token):
    iCurrent = find_next_token(iToken, lObjects)
    iItemCount = 0
    if has_label(iCurrent, lObjects):
        while iItemCount < 2:
            if is_item(lObjects, iCurrent):
                if iItemCount == 0:
                    assign_token(lObjects, iCurrent, label_token)
                if iItemCount == 1:
                    assign_token(lObjects, iCurrent, colon_token)
                iItemCount += 1
            iCurrent += 1
        return iCurrent
    return iToken


def print_debug(sTitle, iStart, iEnd, lObjects):
    print('--> ' + sTitle)
    sOutput = ''
    for iIndex in range(iStart, iEnd + 1):
        print(f'{iIndex} | {lObjects[iIndex]}')
        sOutput += (lObjects[iIndex].get_value())
    print(sOutput)


def print_next_token(iObject, lObjects):
    iCurrent = find_next_token(iObject, lObjects)
    iLine = calculate_line_number(iObject, lObjects)
    print(f'{iLine} | {iCurrent} | {lObjects[iCurrent].get_value()}')


def print_token(iObject, lObjects):
    iLine = calculate_line_number(iObject, lObjects)
    print(f'{iLine} | {iObject} | {lObjects[iObject].get_value()} | {lObjects[iObject]}')


def print_line(lObjects, iStart):
    iIndex = iStart
    sOutput = ''
    while not isinstance(lObjects[iIndex], parser.carriage_return):
        sOutput += lObjects[iIndex].get_value()
        iIndex += 1
    print(sOutput)


def print_lines(lObjects):
    sOutput = ''
    for oObject in lObjects:
        sOutput += oObject.get_value()
    print(sOutput)


def token_is_semicolon(iObject, lObjects):
    if object_value_is(lObjects, iObject, ';'):
        return True
    return False


def token_is_comma(iObject, lObjects):
    if object_value_is(lObjects, iObject, ','):
        return True
    return False


def token_is_open_parenthesis(iObject, lObjects):
    if object_value_is(lObjects, iObject, '('):
        return True
    return False


def token_is_close_parenthesis(iObject, lObjects):
    if object_value_is(lObjects, iObject, ')'):
        return True
    return False


def token_is_assignment_operator(iObject, lObjects):
    if object_value_is(lObjects, iObject, '<='):
        return True
    return False


def increment_token_count(iToken):
    return iToken + 1


def is_next_token(sToken, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    if object_value_is(lObjects, iCurrent, sToken):
        return True
    return False


def is_next_token_one_of(lTokens, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    if lObjects[iCurrent].get_value().lower() in lTokens:
        return True
    return False


def detect_subelement_until(sToken, element, iToken, lObjects):
    iCurrent = iToken
    while not is_next_token(sToken, iCurrent, lObjects):
        iLast = iCurrent
        iCurrent = element.detect(iCurrent, lObjects)
        if iLast == iCurrent:
            return iCurrent
    return iCurrent


def classify_subelement_until(sToken, element, iToken, lObjects):
    iCurrent = iToken
    while not is_next_token(sToken, iCurrent, lObjects):
        iCurrent = element.classify(iCurrent, lObjects)
    return iCurrent


def calculate_line_number(iToken, lObjects):
    iReturn = 1
    for iIndex in range(0, iToken):
        if isinstance(lObjects[iIndex], parser.carriage_return):
            iReturn += 1
    return iReturn


def calculate_column(iToken, lObjects):
    iColumn = 0
    for iCarriageReturn in range(iToken, 0, -1):
        if isinstance(lObjects[iCarriageReturn], parser.carriage_return):
            break;
    for iIndex in range(iCarriageReturn + 1, iToken):
        iColumn += len(lObjects[iIndex].get_value())
    iColumn += 1
    return iColumn


def print_error_message(sToken, token, iToken, lObjects):
    sFoundToken = lObjects[iToken].get_value()
    iLine = calculate_line_number(iToken, lObjects)
    iColumn = calculate_column(iToken, lObjects)
    sModuleName = extract_module_name(token)

    sErrorMessage = '\n'
    sErrorMessage += f'Error: Unexpected token detected while parsing {sModuleName} @ Line {iLine}, Column {iColumn} in file {lObjects[0].get_filename()}'
    sErrorMessage += '\n'
    sErrorMessage += f'       Expecting : {sToken}'
    sErrorMessage += '\n'
    sErrorMessage += f'       Found     : {sFoundToken}'
    sErrorMessage += '\n'

    raise exceptions.ClassifyError(sErrorMessage)


def extract_module_name(token):
    return token.__module__.split('.')[-1]


def keyword_found(sKeyword, iToken, lObjects):
    if find_in_next_n_tokens(':', 2, iToken, lObjects):
        if find_in_next_n_tokens(sKeyword, 3, iToken, lObjects):
            return True
        else:
            return False
    if is_next_token(sKeyword, iToken, lObjects):
        return True
    return False


def is_next_token_in_list(lUntils, iToken, lObjects):
    iCurrent = find_next_token(iToken, lObjects)
    if lObjects[iCurrent].get_value().lower() in lUntils:
        return True
    return False


def combine_two_token_class_lists(lToi_a, lToi_b):
    '''
    Takes two lists that contain Token classes and merges them into a single list.
    The final list is ordered by the iStartIndex attribute on the Token classes.
    '''
    if len(lToi_a) == 0:
        return lToi_b
    if len(lToi_b) == 0:
        return lToi_a
    lReturn = lToi_a.copy()
    for oToi in lToi_b:
        bInserted = False
        for iIndex, oReturn in enumerate(lReturn):
            if oToi.iStartIndex < oReturn.iStartIndex:
                lReturn.insert(iIndex, oToi)
                bInserted = True
                break
        if not bInserted:
            lReturn.append(oToi)
    return lReturn


def does_length_of_tokens_exceed(lObjects, iLength):

    iTotalLength = 0
    for oObject in lObjects:
        iTotalLength += len(oObject.get_value())
    if iTotalLength > iLength:
        return True
    return False


def remove_carriage_returns_from_token_list(lTokens):
    lMyTokens = []
    for oToken in lTokens:
        if isinstance(oToken, parser.carriage_return):
            continue
        lMyTokens.append(oToken)
    return lMyTokens


def remove_comments_from_token_list(lTokens):
    lMyTokens = []
    for oToken in lTokens:
        if isinstance(oToken, parser.comment):
            continue
        lMyTokens.append(oToken)
    return lMyTokens


def remove_consecutive_whitespace_tokens(lTokens):
    lMyTokens = []
    for iToken, oToken in enumerate(lTokens):
        if iToken == 0:
            lMyTokens.append(oToken)
        else:
            if isinstance(oToken, parser.whitespace) and isinstance(lTokens[iToken - 1], parser.whitespace):
                continue
            else:
                lMyTokens.append(oToken)
    return lMyTokens


def remove_whitespace_from_token_list(lTokens):
    lMyTokens = []
    for oToken in lTokens:
        if isinstance(oToken, parser.whitespace):
            continue
        lMyTokens.append(oToken)
    return lMyTokens


def remove_comment_at_end_of_token_list(lTokens):
    if isinstance(lTokens[-1], parser.comment):
        return lTokens[:-1]
    return lTokens


def does_token_type_match(oToken, oType):
    if isinstance(oToken, oType):
        return True
    return False


def remove_trailing_whitespace(lTokens):
    lTokens.reverse()
    for iIndex in range(0, len(lTokens)):
        if token_is_whitespace(lTokens[iIndex]):
            continue
        else:
            lMyTokens = lTokens[iIndex:]
            lMyTokens.reverse()
            return lMyTokens
    else:
        return lTokens


def remove_trailing_whitespace_and_comments(lTokens):
    lTokens.reverse()
    for iIndex in range(0, len(lTokens)):
        if token_is_whitespace_or_comment(lTokens[iIndex]):
            continue
        else:
            lMyTokens = lTokens[iIndex:]
            lMyTokens.reverse()
            return lMyTokens
    else:
        return lTokens


def remove_leading_whitespace_and_comments(iToken, lTokens):
    for iIndex in range(0, len(lTokens)):
        if token_is_whitespace_or_comment(lTokens[iIndex]):
            continue
        else:
            return iToken + iIndex + 1, lTokens[iIndex:]
    else:
        return iToken, lTokens


def token_is_whitespace_or_comment(oToken):
    if isinstance(oToken, parser.whitespace) or \
       isinstance(oToken, parser.carriage_return) or \
       isinstance(oToken, parser.comment) or \
       isinstance(oToken, parser.blank_line) or \
       isinstance(oToken, parser.preprocessor):
        return True
    else:
        return False


def token_is_whitespace(oToken):
    if isinstance(oToken, parser.whitespace) or \
       isinstance(oToken, parser.carriage_return) or \
       isinstance(oToken, parser.blank_line) or \
       isinstance(oToken, parser.preprocessor):
        return True
    else:
        return False


def increment_line_number(iLine, oToken):
    if isinstance(oToken, parser.carriage_return):
        return iLine + 1
    return iLine


def decrement_line_number(iLine, oToken):
    if isinstance(oToken, parser.carriage_return):
        return iLine - 1
    return iLine


def count_carriage_returns(lTokens):
    iReturn = 0
    for oToken in lTokens:
        iReturn = increment_line_number(iReturn, oToken)
    return iReturn


def increment_comment_counter(iComment, oToken):
    if isinstance(oToken, parser.comment):
        return iComment + 1
    return iComment


def find_carriage_return(lTokens, iToken=0):
    for iIndex in range(iToken, len(lTokens)):
        if isinstance(lTokens[iIndex], parser.carriage_return):
            return iIndex
    return None


def count_token_types_in_list_of_tokens(oType, lTokens):
    iReturn = 0
    for oToken in lTokens:
        if isinstance(oToken, oType):
            iReturn += 1
    return iReturn


def does_token_type_exist_in_list_of_tokens(oType, lTokens):
    if count_token_types_in_list_of_tokens(oType, lTokens) == 0:
        return False
    return True


def get_toi_parameters(oToi):
    return oToi.get_line_number(), oToi.get_tokens()


def does_token_start_line(iToken, lTokens):
    if isinstance(lTokens[iToken - 1], parser.whitespace) and isinstance(lTokens[iToken - 2], parser.carriage_return):
        return True
    if isinstance(lTokens[iToken - 1], parser.carriage_return):
        return True
    return False


def convert_token_list_to_string(lTokens):
    sReturn = ''
    for oToken in lTokens:
        sReturn += oToken.get_value()
    return sReturn


def fix_blank_lines(lTokens):
    lReturn = []
    for iToken, oToken in enumerate(lTokens):
        try:
            if isinstance(oToken, parser.carriage_return) and isinstance(lTokens[iToken + 1], parser.carriage_return):
                lReturn.append(oToken)
                lReturn.append(parser.blank_line())
                continue
        except IndexError:
            pass
        try:
            if isinstance(lTokens[iToken - 1], parser.carriage_return) and isinstance(oToken, parser.whitespace) and isinstance(lTokens[iToken + 1], parser.carriage_return):
                lReturn.append(parser.blank_line())
                continue
        except IndexError:
            pass
        lReturn.append(oToken)

    return lReturn


def fix_trailing_whitespace(lTokens):
    lReturn = []
    for iToken, oToken in enumerate(lTokens):
        try:
            if isinstance(oToken, parser.carriage_return) and isinstance(lTokens[iToken - 1], parser.whitespace):
                lReturn.pop()
                lReturn.append(oToken)
                continue
        except IndexError:
            pass
        lReturn.append(oToken)

    return lReturn


def is_whitespace(oObject):
    if isinstance(oObject, parser.carriage_return):
        return True
    if isinstance(oObject, parser.blank_line):
        return True
    if isinstance(oObject, parser.comment):
        return True
    if isinstance(oObject, parser.whitespace):
        return True
    if isinstance(oObject, parser.preprocessor):
        return True
    return False


def read_vhdlfile(sFileName):
    try:
        lLines = []
        with open(sFileName, encoding='utf-8') as oFile:
            for sLine in oFile:
                lLines.append(sLine)
        return lLines, None
    except UnicodeDecodeError:
        lLines = []
        with open(sFileName, encoding="ISO-8859-1") as oFile:
            for sLine in oFile:
                lLines.append(sLine)
        return lLines, None
    except OSError as e:
        return [], e


def is_token_at_end_of_line(iToken, lTokens):
    iMyToken = iToken + 1
    if are_next_consecutive_token_types([parser.carriage_return], iMyToken, lTokens):
        return True
    if are_next_consecutive_token_types([parser.whitespace, parser.carriage_return], iMyToken, lTokens):
        return True
    if are_next_consecutive_token_types([parser.comment, parser.carriage_return], iMyToken, lTokens):
        return True
    if are_next_consecutive_token_types([parser.whitespace, parser.comment, parser.carriage_return], iMyToken, lTokens):
        return True
    return False


def find_next_token_with_value(iToken, sValue, lTokens):
    for iIndex, oToken in enumerate(lTokens[iToken::]):
        if oToken.get_value() == sValue:
            return iToken + iIndex
    return None


def all_assignments_inside_parenthesis(iToken, sStop, lTokens):
    iStop = find_next_token_with_value(iToken, sStop, lTokens)
    iParen = 0
    for iIndex in range(iToken, iStop + 1):
        iParen = update_paren_counter(iIndex, lTokens, iParen)
        if token_is_assignment_operator(iIndex, lTokens) and iParen == 0:
            return False
    return True


def update_paren_counter(iToken, lTokens, iCounter):
    if token_is_open_parenthesis(iToken, lTokens):
        return iCounter + 1
    if token_is_close_parenthesis(iToken, lTokens):
        return iCounter - 1
    return iCounter


def assignment_operator_found(iToken, lObjects):
    if find_in_range('<=', iToken, ';', lObjects):
        if all_assignments_inside_parenthesis(iToken, ';', lObjects):
            return False
        return True
    return False


def assign_special_tokens(lObjects, iCurrent, oType):
    sValue = lObjects[iCurrent].get_value()
    if sValue == ')':
        assign_token(lObjects, iCurrent, parser.close_parenthesis)
    elif sValue == '(':
        assign_token(lObjects, iCurrent, parser.open_parenthesis)
    elif sValue == '-':
        if isinstance(lObjects[iCurrent - 1], exponent.e_keyword):
            assign_token(lObjects, iCurrent, exponent.minus_sign)
        else:
            assign_token(lObjects, iCurrent, parser.todo)
    elif sValue == '+':
        if isinstance(lObjects[iCurrent - 1], exponent.e_keyword):
            assign_token(lObjects, iCurrent, exponent.plus_sign)
        else:
            assign_token(lObjects, iCurrent, parser.todo)
    elif sValue == '*':
        assign_token(lObjects, iCurrent, parser.todo)
    elif sValue == '**':
        assign_token(lObjects, iCurrent, parser.todo)
    elif sValue == '/':
        assign_token(lObjects, iCurrent, parser.todo)
    elif sValue.lower() == 'downto':
        assign_token(lObjects, iCurrent, direction.downto)
    elif sValue.lower() == 'to':
        assign_token(lObjects, iCurrent, direction.to)
    elif sValue.lower() == 'std_logic_vector':
        assign_token(lObjects, iCurrent, types.std_logic_vector)
    elif sValue.lower() == 'std_ulogic_vector':
        assign_token(lObjects, iCurrent, types.std_ulogic_vector)
    elif sValue.lower() == 'std_ulogic':
        assign_token(lObjects, iCurrent, types.std_ulogic)
    elif sValue.lower() == 'std_logic':
        assign_token(lObjects, iCurrent, types.std_logic)
    elif sValue.lower() == 'integer':
        assign_token(lObjects, iCurrent, types.integer)
    elif sValue.lower() == 'signed':
        assign_token(lObjects, iCurrent, types.signed)
    elif sValue.lower() == 'unsigned':
        assign_token(lObjects, iCurrent, types.unsigned)
    elif sValue.lower() == 'natural':
        assign_token(lObjects, iCurrent, types.natural)
    elif sValue.lower() == 'others':
        assign_token(lObjects, iCurrent, choice.others_keyword)
    elif sValue.lower() == '=>':
        assign_token(lObjects, iCurrent, element_association.assignment)
    elif sValue.lower() == 'e':
        if lObjects[iCurrent + 1].get_value().isdigit() or lObjects[iCurrent + 1].get_value() == '-' or lObjects[iCurrent + 1].get_value() == '+':
            assign_token(lObjects, iCurrent, exponent.e_keyword)
        else:
            assign_token(lObjects, iCurrent, oType)
    elif exponent_detected(lObjects, iCurrent):
        assign_token(lObjects, iCurrent, exponent.integer)
    else:
        assign_token(lObjects, iCurrent, oType)


def exponent_detected(lObjects, iCurrent):
    if isinstance(lObjects[iCurrent - 1], exponent.e_keyword):
        return True
    if isinstance(lObjects[iCurrent - 1], exponent.plus_sign):
        return True
    if isinstance(lObjects[iCurrent - 1], exponent.minus_sign):
        return True
    return False


def classify_predefined_types(lObjects, iCurrent):
    if not isinstance(lObjects[iCurrent], parser.todo):
        return
    if lObjects[iCurrent].get_value().lower() in predefined_attribute.values:
        if lObjects[iCurrent].get_value().lower() == 'event':
            assign_token(lObjects, iCurrent, predefined_attribute.event_keyword)
        else:
            assign_token(lObjects, iCurrent, predefined_attribute.keyword)




from vsg.token import conditional_variable_assignment as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import conditional_expressions


def detect(iToken, lObjects):
    '''
    conditional_variable_assignment ::=
        target := conditional_expressions ;
    '''

    if utils.is_next_token_one_of(['when', 'if', 'elsif', 'else'], iToken, lObjects):
        return False
    if utils.find_in_range(':=', iToken, ';', lObjects):
        if not utils.find_in_range('with', iToken, ';', lObjects):
            if utils.find_in_range('when', iToken, ';', lObjects):
                return True
    return False


def classify(iToken, lObjects):

    iCurrent = utils.assign_tokens_until(':=', token.target, iToken, lObjects)
    iCurrent = utils.assign_next_token_required(':=', token.assignment, iCurrent, lObjects)

    iCurrent = conditional_expressions.classify_until([';'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import token

from vsg.rules import n_spaces_before_and_after_tokens

lTokens = []
lTokens.append(token.adding_operator.plus)
lTokens.append(token.adding_operator.minus)
lTokens.append(token.multiplying_operator.star)
lTokens.append(token.multiplying_operator.slash)
lTokens.append(token.miscellaneous_operator.double_star)


class rule_011(n_spaces_before_and_after_tokens):
    '''
    This rule checks for at least a single space before and after math operators +, -, /, \* and \*\*.

    **Violation**

    .. code-block:: vhdl

       a <= b+c;
       a <= b-c;
       a <= b/c;
       a <= b*c;
       a <= b**c;
       a <= (b+c)-(d-e);

    **Fix**

    .. code-block:: vhdl

       a <= b + c;
       a <= b - c;
       a <= b / c;
       a <= b * c;
       a <= b ** c;
       a <= (b + c) - (d - e);
    '''

    def __init__(self):
        n_spaces_before_and_after_tokens.__init__(self, 'whitespace', '011', 1, lTokens, bNIsMinimum=True)
        self.solution = 'Ensure at least a single space before and after math operator.'




from vsg.rules import insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token as Rule

from vsg.token import record_type_definition as token
from vsg.token import full_type_declaration as ft_token

oInsertToken = token.record_type_simple_name
oAnchorToken = ft_token.semicolon
oLeftToken = token.end_record_keyword
oRightToken = ft_token.semicolon
oValueToken = ft_token.identifier


class rule_005(Rule):
    '''
    This rule checks for the optional simple name in the **end record** statement.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record t_record;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '005', oInsertToken, oAnchorToken, oLeftToken, oRightToken, oValueToken)
        self.subphase = 1
        self.solution = 'record type simple name'
        self.groups.append('structure::optional')




from vsg.token import interface_package_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import identifier
from vsg.vhdlFile.classify import interface_package_generic_map_aspect


def detect(iToken, lObjects):
    '''
    interface_package_declaration ::=
        package identifier is
            new *uninstantiated_package*_name interface_package_generic_map_aspect
    '''
    if utils.is_next_token('package', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('package', token.package_keyword, iToken, lObjects)

    iCurrent = identifier.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('new', token.new_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token(token.uninstantiated_package_name, iCurrent, lObjects)

    iCurrent = interface_package_generic_map_aspect.detect(iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.entity_declaration.end_keyword)


class rule_009(token_indent):
    '''
    This rule checks the indent of the **end** keyword.

    **Violation**

    .. code-block:: vhdl

         wr_en : in    std_logic;
         rd_en : in    std_logic
       );
         end entity fifo;

    **Fix**

    .. code-block:: vhdl

           wr_en : in    std_logic;
           rd_en : in    std_logic
         );
       end entity fifo;
    '''

    def __init__(self):
        token_indent.__init__(self, 'entity', '009', lTokens)




from vsg import parser
from vsg import token

from vsg.vhdlFile.extract import tokens


def get_association_elements_between_tokens(oStart, oEnd, lAllTokens, oTokenMap):
    lReturn = []
    lStartIndexes, lEndIndexes = oTokenMap.get_token_pair_indexes(oStart, oEnd)

    for iStart, iEnd in zip(lStartIndexes, lEndIndexes):
        iLine = oTokenMap.get_line_number_of_index(iStart)
        lTokens = lAllTokens[iStart:iEnd + 1]

        bStore = False
        iLineNumber = None
        lTemp = []
        for iToken, oToken in enumerate(lTokens):
            if isinstance(oToken, token.association_element.formal_part):
                bStore = True
                iStartIndex = iToken + iStart
                iLineNumber = iLine
            if isinstance(oToken, token.association_element.actual_part) and not bStore:
                bStore = True
                iStartIndex = iToken + iStart
                iLineNumber = iLine

            if bStore:
               lTemp.append(oToken)

            if isinstance(oToken, token.association_list.comma):
                lReturn.append(tokens.New(iStartIndex, iLineNumber, lTemp))
                lTemp = []
                bStore = False

            if isinstance(oToken, parser.carriage_return):
                iLine +=1

        if len(lTemp) > 0:
            lReturn.append(tokens.New(iStartIndex, iLineNumber, lTemp))

    return lReturn




from vsg.rules import blank_line_above_line_starting_with_token as rule


class Rule(rule):
    '''
    Checks for a blank line above a line starting with a given token when it is between tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    token: token object type list
       token object that a blank line above should appear
    '''

    def __init__(self, name, identifier, lTokens, lAllowTokens=None):
        rule.__init__(self, name=name, identifier=identifier, lTokens=lTokens, lAllowTokens=lAllowTokens)
        self.lBetweenTokenPairs = None

    def _get_tokens_of_interest(self, oFile):
        if self.style == 'require_blank_line':
            return oFile.get_line_above_line_starting_with_token(self.lTokens, bIncludeComments=False)
        elif self.style == 'no_blank_line':
            return oFile.get_blank_lines_above_line_starting_with_token_when_between_tokens(self.lTokens, self.lBetweenTokenPairs)




from vsg import parser
from vsg import violation

from vsg.token import process_statement as token

from vsg.rule_group import blank_line

from vsg.rules import utils as rules_utils

from vsg.vhdlFile import utils


class rule_027(blank_line.Rule):
    '''
    This rule checks for blank lines above the **begin** keyword if a declarative item exists.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is

         -- Keep track of the number of words in the FIFO
         variable word_count : integer;
       begin

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is

         -- Keep track of the number of words in the FIFO
         variable word_count : integer;

       begin
    '''

    def __init__(self):
        blank_line.Rule.__init__(self, 'process', '027')
        self.phase = 3
        self.style = 'require_blank_line'
        self.configuration.append('style')

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(token.process_keyword, token.begin_keyword)

    def _analyze(self, lToi):
        if self.style == 'require_blank_line':
            _analyze_require_blank_line(self, lToi)
        elif self.style == 'no_blank_line':
            _analyze_no_blank_line(self, lToi)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        if dAction['action'] == 'Insert':
            rules_utils.insert_carriage_return(lTokens, dAction['index'])
            rules_utils.insert_blank_line(lTokens, dAction['index'])
            oViolation.set_tokens(lTokens)
        elif dAction['action'] == 'Remove':
            iStart = dAction['start']
            iEnd = dAction['end']
            lNewTokens = lTokens[:iStart]
            lNewTokens.extend(lTokens[iEnd:])
            oViolation.set_tokens(lNewTokens)


def _analyze_require_blank_line(self, lToi):
    sSolution = 'Insert blank line above *begin* keyword'
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        iLine = oToi.get_line_number() + utils.count_carriage_returns(lTokens)
        lTokens.reverse()
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.begin_keyword, token.is_keyword], 0, lTokens):
            continue
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.begin_keyword, token.close_parenthesis], 0, lTokens):
            continue
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.begin_keyword, token.process_keyword], 0, lTokens):
            continue
        if utils.are_next_consecutive_token_types([token.begin_keyword, parser.whitespace, parser.carriage_return, parser.blank_line], 0, lTokens):
            continue
        if utils.are_next_consecutive_token_types([token.begin_keyword, parser.carriage_return, parser.blank_line], 0, lTokens):
            continue
        dAction = {}
        dAction['action'] = 'Insert'
        if isinstance(lTokens[1], parser.whitespace):
            dAction['index'] = len(lTokens) - 2
        else:
            dAction['index'] = len(lTokens) - 1
        lTokens.reverse()
        oViolation = violation.New(iLine, oToi, sSolution)
        oViolation.set_action(dAction)
        self.add_violation(oViolation)

def _analyze_no_blank_line(self, lToi):
    sSolution = 'Remove blank line(s) above *begin* keyword'
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        iLine = oToi.get_line_number() + utils.count_carriage_returns(lTokens)
        lTokens.reverse()
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.begin_keyword, token.is_keyword], 0, lTokens):
            continue
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.begin_keyword, token.close_parenthesis], 0, lTokens):
            continue
        if utils.are_next_consecutive_token_types_ignoring_whitespace([token.begin_keyword, token.process_keyword], 0, lTokens):
            continue
        if not utils.are_next_consecutive_token_types([token.begin_keyword, parser.whitespace, parser.carriage_return, parser.blank_line], 0, lTokens) and \
           not utils.are_next_consecutive_token_types([token.begin_keyword, parser.carriage_return, parser.blank_line], 0, lTokens):
            continue
        dAction = {}
        dAction['action'] = 'Remove'

        if isinstance(lTokens[1], parser.whitespace):
            iEnd = len(lTokens) - 2
        else:
            iEnd = len(lTokens) - 3

        for iToken, oToken in enumerate(lTokens):
            if isinstance(oToken, parser.carriage_return):
                if not isinstance(lTokens[iToken + 1], parser.carriage_return):
                    iStart = len(lTokens) - iToken - 2
                    break

        lTokens.reverse()

        dAction['start'] = iStart
        dAction['end'] = iEnd
        oViolation = violation.New(iLine, oToi, sSolution)
        oViolation.set_action(dAction)
        self.add_violation(oViolation)




import vsg

from vsg import config
from vsg import vhdlFile
from vsg import rule_list
from vsg import apply_rules


class New():

    def __init__(self):
        self.identifier = 'notepad++ interface'

    def get_input_arguments(self):
        return InputArguments()

    def fix(self, oInputArguments):
        lText = oInputArguments.text.splitlines()
        iIndex = 0
        sFileName = oInputArguments.filename
        commandLineArguments = command_line_args()
        commandLineArguments.style = oInputArguments.style
        commandLineArguments.configuration = oInputArguments.configuration
        oConfig = config.New(commandLineArguments)

        oResults = Results()

        try:
            oVhdlFile = vhdlFile.vhdlFile(lText, sFileName, None)
        except vsg.exceptions.ClassifyError as e:
            oResults.error = True
            oResults.set_violations(False)
            oResults.set_text(oInputArguments.text)
            sOutput = e.message
            oResults.set_stdout(sOutput)
            return oResults

        oVhdlFile.set_indent_map(oConfig.dIndent)

        oRules = rule_list.rule_list(oVhdlFile, oConfig.severity_list, commandLineArguments.local_rules)

        try:
            apply_rules.configure_rules(oConfig, oRules, oConfig.dIndent, iIndex, sFileName)
        except vsg.exceptions.ConfigurationError as e:
            oResults.error = True
            oResults.set_violations(False)
            oResults.set_text(oInputArguments.text)
            sOutput = e.message
            oResults.set_stdout(sOutput)
            return oResults

        oRules.fix(commandLineArguments.fix_phase, commandLineArguments.skip_phase, oConfig.dFixOnly)

        sOutput = '\n'.join(oVhdlFile.get_lines()[1:])

        oResults.set_text(sOutput)

        oRules.clear_violations()
        oRules.check_rules(
            bAllPhases=commandLineArguments.all_phases,
            lSkipPhase=commandLineArguments.skip_phase,
        )
        sStdOut, sIgnore = oRules.report_violations(commandLineArguments.output_format)
        oResults.set_stdout(sStdOut)

        oResults.set_violations(oRules.violations)

        return oResults


class InputArguments():

    def __init__(self):
        self.filename = str(None)
        self.text = None
        self.style = None
        self.configuration = []

    def set_text(self, sText):
        self.text = sText

    def set_style(self, sText):
        self.style = sText

    def add_configuration(self, sText):
        self.configuration.append(sText)


class Results():

    def __init__(self):
        self.text = None
        self.stdout = None
        self.violations = True
        self.error = False

    def set_text(self, sText):
        self.text = sText

    def get_text(self):
        return self.text

    def set_stdout(self, sText):
        self.stdout = sText

    def get_stdout(self):
        return self.stdout

    def has_violations(self):
        return self.violations

    def set_violations(self, bViolations):
        self.violations = bViolations

    def error_status(self):
        return self.error


class command_line_args():
    def __init__(self):
        self.fix = False
        self.style = None
        self.configuration = None
        self.debug = None
        self.fix_only = None
        self.local_rules = None
        self.fix_phase = 7
        self.junit = None
        self.output_format = 'vsg'
        self.all_phases = False
        self.skip_phase = None




import os


def installing_from_git_repo():
    sPath = os.path.dirname(__file__)
    if os.path.isdir(os.path.join(sPath, '..', '.git')):
        return True
    return False


def create_version_info_file(sVersion, sShaNum):
    sVersionInfoFileName = os.path.join(os.path.dirname(__file__), 'version_info.py')

    lVersionInfo = []
    lVersionInfo.append('sVersion = \'' + str(sVersion) + "'")
    lVersionInfo.append('sShaNum = \'' + str(sShaNum) + "'")

    with open(sVersionInfoFileName, 'w') as oFile:
        for sLine in lVersionInfo:
            oFile.write(sLine + '\n')
    oFile.close()




from vsg.deprecated_rule import Rule


class rule_600(Rule):
    '''
    This rule has been moved to library_500.
    '''

    def __init__(self):
        Rule.__init__(self, 'library', '600')
        self.message.append('Rule ' + self.unique_id + ' has been moved to rule library_500.')



from vsg.token import element_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import element_subtype_definition
from vsg.vhdlFile.classify import identifier_list


def classify(iToken, lObjects):
    '''
    element_declaration ::=
        identifier_list : element_subtype_definition ;
    '''

    iCurrent = identifier_list.classify_until([':'], iToken, lObjects)

    iCurrent = utils.assign_next_token_required(':', token.colon, iCurrent, lObjects)

    iCurrent = element_subtype_definition.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent
